"""This module contains utility functions related to the weighted_tensor module."""

from typing import Tuple, TypeVar, Union

import torch

from ._weighted_tensor import TensorOrWeightedTensor, WeightedTensor

__all__ = [
    "expand_left",
    "expand_right",
    "unsqueeze_right",
    "sum_dim",
    "wsum_dim",
    "wsum_dim_return_weighted_sum_only",
    "wsum_dim_return_sum_of_weights_only",
]
S = TypeVar("S")


def expand_left(
    t: TensorOrWeightedTensor[S], *, shape: Tuple[int, ...]
) -> TensorOrWeightedTensor[S]:
    """Expand shape of tensor at left with provided shape."""
    return t.expand(shape + t.shape)


def expand_right(
    t: TensorOrWeightedTensor[S], *, shape: Tuple[int, ...]
) -> TensorOrWeightedTensor[S]:
    """Expand shape of tensor at right with provided shape."""
    return t.expand(t.shape + shape)


def unsqueeze_right(
    t: TensorOrWeightedTensor[S], *, ndim: int
) -> TensorOrWeightedTensor[S]:
    """
    Adds `ndim` dimensions to tensor, from right-side, without
    copy (useful for right broadcasting which is non-standard).
    """
    # Nota: `unsqueeze_left` is useless since it is automatically done with standard broadcasting
    assert isinstance(ndim, int) and ndim >= 0, f"Can not unsqueeze {ndim} dimensions"
    if ndim == 0:
        return t
    return t.view(t.shape + (1,) * ndim)


def sum_dim(
    x: TensorOrWeightedTensor,
    *,
    fill_value=0,
    dim: Union[None, int, Tuple[int, ...]] = None,
    but_dim: Union[None, int, Tuple[int, ...]] = None,
    **kws,
) -> torch.Tensor:
    """
    Sum dimension(s) of provided tensor (regular or weighted -
    filling with `fill_value` aggregates without any summed weighting if any).
    """
    dim = _get_dim(x, dim=dim, but_dim=but_dim)
    if isinstance(x, WeightedTensor):
        return x.sum(fill_value=fill_value, dim=dim, **kws)
    return x.sum(dim=dim, **kws)


def wsum_dim(
    x: WeightedTensor,
    *,
    fill_value=0,
    dim: Union[None, int, Tuple[int, ...]] = None,
    but_dim: Union[None, int, Tuple[int, ...]] = None,
    **kws,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Sum dimension(s) of provided weighted tensor.

    The weighted tensor is filled with `fill_value` if provided.
    The function returns the sum of weights as well.

    Parameters
    ----------
    x : WeightedTensor
        The weighted tensor on which to compute the sum.
    fill_value : float
        The value to use for filling the weighted tensor.
    dim : int or tuple of int, optional
        The dimension(s) on which to sum.
    but_dim : int or tuple of int, optional
        The dimension(s) to omit when summing.

    Returns
    -------
    weighted_sum : torch.Tensor
        Weighted sum, with totally un-weighted aggregates filled with `fill_value`.
    sum_weights : torch.Tensor
        The sum of weights.
    """
    dim = _get_dim(x, dim=dim, but_dim=but_dim)
    return x.wsum(fill_value=fill_value, dim=dim, **kws)


def wsum_dim_return_weighted_sum_only(
    x: WeightedTensor,
    *,
    fill_value=0,
    dim: Union[None, int, Tuple[int, ...]] = None,
    but_dim: Union[None, int, Tuple[int, ...]] = None,
    **kws,
) -> torch.Tensor:
    """
    Sum dimension(s) of provided weighted tensor like `wsum_dim` but
    only return the weighted sum and not the sum of weights.

    Parameters
    ----------
    x : WeightedTensor
        The weighted tensor on which to compute the sum.
    fill_value : float
        The value to use for filling the weighted tensor.
    dim : int or tuple of int, optional
        The dimension(s) on which to sum.
    but_dim : int or tuple of int, optional
        The dimension(s) to omit when summing.

    Returns
    -------
    torch.Tensor :
        Weighted sum, with totally un-weighted aggregates filled with `fill_value`.
    """
    return wsum_dim(x, fill_value=fill_value, dim=dim, but_dim=but_dim, **kws)[0]


def wsum_dim_return_sum_of_weights_only(
    x: WeightedTensor,
    *,
    fill_value=0,
    dim: Union[None, int, Tuple[int, ...]] = None,
    but_dim: Union[None, int, Tuple[int, ...]] = None,
    **kws,
) -> torch.Tensor:
    """
    Sum dimension(s) of provided weighted tensor like `wsum_dim` but
    only return the sum of weights and not the weighted sum.

    Parameters
    ----------
    x : WeightedTensor
        The weighted tensor on which to compute the sum.
    fill_value : float
        The value to use for filling the weighted tensor.
    dim : int or tuple of int, optional
        The dimension(s) on which to sum.
    but_dim : int or tuple of int, optional
        The dimension(s) to omit when summing.

    Returns
    -------
    torch.Tensor :
        The sum of weights.
    """
    return wsum_dim(x, fill_value=fill_value, dim=dim, but_dim=but_dim, **kws)[1]


def _get_dim(
    x: TensorOrWeightedTensor,
    *,
    dim: Union[None, int, Tuple[int, ...]] = None,
    but_dim: Union[None, int, Tuple[int, ...]] = None,
) -> Union[int, Tuple[int, ...]]:
    if (dim is not None) and (but_dim is not None):
        raise ValueError("`dim` and `but_dim` should not be both defined.")
    if but_dim is not None:
        ndim = x.ndim
        if isinstance(but_dim, int):
            but_dim = {but_dim}
        but_dim = {i if i >= 0 else ndim + i for i in but_dim}
        assert all(i >= 0 for i in but_dim), but_dim
        dim = tuple(i for i in range(ndim) if i not in but_dim)
    elif dim is None:
        # full summation by default
        dim = ()
    return dim



from dataclasses import dataclass

__all__ = ["constants"]


@dataclass(frozen=True)
class LeaspyConstants:
    # The real infinity is not used because there are multiplication by zero that creates errors (nan)
    INFINITY = float(10**307)


constants = LeaspyConstants()



import json
import warnings
from abc import ABC
from enum import Enum

import numpy as np
import pandas as pd
import torch
from scipy.stats import beta

from leaspy.algo import AlgorithmSettings
from leaspy.algo.base import AlgorithmType, BaseAlgorithm
from leaspy.algo.simulate.base import BaseSimulationAlgorithm
from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.io.data.data import Data
from leaspy.io.outputs import IndividualParameters
from leaspy.io.outputs.result import Result
from leaspy.models import BaseModel, McmcSaemCompatibleModel


class VisitType(str, Enum):
    """Enum for different types of visit simulations.

    Attributes
    ----------
    DATAFRAME : :obj:`str`
        Represents visits defined by a DataFrame containing visit times.
    RANDOM : :obj:`str`
        Represents visits generated randomly based on specified parameters.
    """

    DATAFRAME = "dataframe"  # Dataframe of visits
    RANDOM = "random"  # Random spaced visits


class SimulationAlgorithm(BaseSimulationAlgorithm):
    """
    SimulationAlgorithm class for simulating longitudinal data using a Leaspy model.
    This class extends the BaseSimulationAlgorithm and implements the simulation logic
    based on the provided model and settings.

    Attributes
    ----------
    name : :obj:`str`
        The name of the algorithm.

    family : :class:`~leaspy.algo.base.AlgorithmType`
        The type of algorithm, which is AlgorithmType.SIMULATE.

    PARAM_REQUIREMENTS : :obj:`dict`
        A dictionary defining the required parameters for different visit types.
        It specifies the expected types and constraints for each parameter.

        for visit_type "dataframe":
            - 'df_visits': :obj:`pandas.DataFrame`
                DataFrame of visits, with a column "ID" and a column 'TIME'.
                TIME and number of visits for each simulated patients (with specified ID)
                are given by a dataframe in dict_param.

        for visit_type "random":
            - 'patient_number': :obj:`int`
                Number of patients.
            - 'first_visit_mean': :obj:`float`
                Mean of the first visit TIME.
            - 'first_visit_std': :obj:`float`
                Standard deviation of the first visit TIME.
            - 'time_follow_up_mean': :obj:`float`
                Mean of the follow-up TIME.
            - 'time_follow_up_std': :obj:`float`
                Standard deviation of the follow-up TIME.
            - 'distance_visit_mean': :obj:`float`
                Mean interval between two visits.
            - 'distance_visit_std': :obj:`float`
               Standard deviation of the intervals between visits.
                Time interval between 2 visits is drawn in a normal distribution N(distance_visit_mean, distance_visit_std),
                thus setting distance_visit_std to 0 enable to simulate regularly spaced visits.
            - 'min_spacing_between_visits': :obj:`float`
                Minimum interval between two visits. This interval has to be in the same unit as the TIME column.
                If two visits are closer than this value, the second visit will be removed from the dataset.
                This is used to avoid too close visits in the simulated dataset.
                Default is 1/365 (1 day).

    """

    name: str = "simulate"
    family: AlgorithmType = AlgorithmType.SIMULATE

    _PARAM_REQUIREMENTS = {
        "dataframe": [
            ("df_visits", pd.DataFrame),
        ],
        "random": [
            ("patient_number", int),
            ("first_visit_mean", (int, float)),
            ("first_visit_std", (int, float)),
            ("time_follow_up_mean", (int, float)),
            ("time_follow_up_std", (int, float)),
            ("distance_visit_mean", (int, float)),
            ("distance_visit_std", (int, float)),
        ],
    }

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.features = settings.parameters["features"]
        self.visit_type = settings.parameters["visit_parameters"]["visit_type"]
        self._set_param_study(settings.parameters["visit_parameters"])
        self._validate_algo_parameters()

    def _check_features(self):
        """Check if the features are valid.

        This method checks if the features are provided as a list of strings.

        Raises
        ------
        LeaspyAlgoInputError
            If the features are not a list or if any of the features is not a string.
        """

        if not isinstance(self.features, list):
            raise LeaspyAlgoInputError(
                f"Features need to a be a list and not : {type(self.features).__name__}"
            )
        if len(self.features) == 0:
            raise LeaspyAlgoInputError("Features can't be empty")

        for i, feature in enumerate(self.features):
            if not isinstance(feature, str):
                raise LeaspyAlgoInputError(
                    f"Invalid feature at position {i}: need to be a string. "
                    f"And not : {type(feature).__name__}"
                )
            if not feature.strip():
                raise LeaspyAlgoInputError(f"Empty feature at the position {i}")

    def _check_params(self, requirements):
        """Check if the parameters are valid.

        This method checks if the parameters in the `param_study` dictionary match the expected types
        and constraints defined in the `requirements` list.

        Parameters
        ----------
        requirements :obj:`list`
            A list of tuples, where each tuple contains a parameter name and its expected type(s).

        Raises
        ------
        LeaspyAlgoInputError
            If any parameter is missing, has an invalid type, or has an invalid value.
        """

        missing_params = []
        type_errors = []
        value_errors = []

        for param, expected_types in requirements:
            if param not in self.param_study:
                missing_params.append(param)
                continue
            value = self.param_study[param]
            if not isinstance(value, expected_types):
                type_names = (
                    [t.__name__ for t in expected_types]
                    if isinstance(expected_types, tuple)
                    else expected_types.__name__
                )
                type_errors.append(
                    f"Parameter '{param}': Expected type {type_names}, given {type(value).__name__}"
                )
            if param == "patient_number" and value <= 0:
                value_errors.append(
                    "Patient number (patient_number) need to be a positive integer"
                )

            if param.endswith("_std") and value < 0:
                value_errors.append(f"Standard deviation ({param}) can't be negative")

        if "min_spacing_between_visits" in self.param_study:
            value = self.param_study["min_spacing_between_visits"]
            if not isinstance(value, (int, float)):
                type_errors.append(
                    "Parameter 'min_spacing_between_visits': Expected type int or float, "
                    f"given {type(value).__name__}"
                )
            if value < 0:
                value_errors.append(
                    "Parameter 'min_spacing_between_visits' cannot be negative"
                )

        errors = []
        if missing_params:
            errors.append(f"Missing parameters : {', '.join(missing_params)}")
        if type_errors:
            errors.append("Type problems :\n- " + "\n- ".join(type_errors))
        if value_errors:
            errors.append("Invalid value :\n- " + "\n- ".join(value_errors))
        if errors:
            raise LeaspyAlgoInputError("\n".join(errors))

    def _check_logistic_model(self, model: McmcSaemCompatibleModel):
        """Check if the model is a logistic model.

        This method checks if the model type is 'logistic' and raises an error if not.
        Parameters
        ----------
        model : :class:~.models.abstract_model.McmcSaemCompatibleModel
            A Leaspy model object previously trained on longitudinal data.
        Raises
        ------
        LeaspyAlgoInputError
            If the model type is not 'logistic'.
        """
        if model.__class__.__name__ != "LogisticModel":
            raise LeaspyAlgoInputError(
                "The model type should be 'logistic' (LogisticModel) for simulation."
            )

    def _validate_algo_parameters(self):
        """Validate the algorithm parameters.

        This method checks the visit type, features, and parameters of the algorithm.

        Raises
        ------
        LeaspyAlgoInputError
            If the visit type is invalid, if the features are not a list of strings,
            or if the parameters do not meet the expected requirements.
        """
        self._check_features()

        requirements = self._PARAM_REQUIREMENTS.get(self.visit_type)
        if not requirements:
            raise LeaspyAlgoInputError(
                f"No configuration for this type of visit '{self.visit_type}'"
            )

        self._check_params(requirements)

        if self.visit_type == VisitType.DATAFRAME:
            df = self.param_study["df_visits"]
            if "ID" not in df.columns or "TIME" not in df.columns:
                raise LeaspyAlgoInputError(
                    "Dataframe needs to have columns 'ID' and 'TIME'"
                )

            if df["TIME"].isnull().any():
                raise LeaspyAlgoInputError("Dataframe has null value in column TIME")

        if self.visit_type == VisitType.RANDOM:
            if (
                self.param_study["distance_visit_mean"] <= 0
                and self.param_study["distance_visit_std"] <= 0
            ):
                raise LeaspyAlgoInputError(
                    "Distance visit mean (distance_visit_mean) and distance visit std need to be positive"
                )

    ## --- SET PARAMETERS ---
    # def _save_parameters(self, model, path_save):  # TODO
    #     total_params = {"study": self.param_study, "model": model.parameters}
    #     with open(f"{path_save}params_simulated.json", "w") as outfile:
    #         json.dump(total_params, outfile)

    def _set_param_study(self, dict_param: dict) -> None:
        """Set parameters related to the study based on visit type.

        This function initializes the `param_study` attribute with relevant
        parameters depending on the visit type of the object. It handles
        three different visit types: 'dataframe' and 'random',
        each requiring a different set of input parameters.

        Parameters
        ----------
        dict_param : :obj:`dict`
            Dictionary containing parameters required for the study. The
            expected keys vary depending on the visit type:

            - If `visit_type` is "dataframe":
                - 'df_visits' : :obj:`pandas.DataFrame`
                    DataFrame of visits, with a column "ID" and a column 'TIME'.
                TIME and number of visits for each simulated patients (with specified ID)
                are given by a dataframe in dict_param.

            - If `visit_type` is "random":
                - 'patient_number' : :obj:`int`
                    Number of patients.
                - 'first_visit_mean' : :obj:`float`
                    Mean of the first visit TIME.
                - 'first_visit_std' : :obj:`float`
                    Standard deviation of the first visit TIME.
                - 'time_follow_up_mean' : :obj:`float`
                    Mean of the follow-up TIME.
                - 'time_follow_up_std' : :obj:`float`
                    Standard deviation of the follow-up TIME.
                - 'distance_visit_mean' : :obj:`float`
                    Mean of distance_visits: mean time delta between two visits.
                - 'distance_visit_std' : :obj:`float`
                    Standard deviation of distance_visits: std time delta between two visits.
                Time delta between 2 visits is drawn in a normal distribution N(distance_visit_mean, distance_visit_std),
                thus setting distance_visit_std to 0 enable to simulate regularly spaced visits.
                - 'min_spacing_between_visits' : :obj:`float`
                    Minimum delta between visits. This delta has to be in the same unit as the TIME column.
                    If two visits are closer than this value, the second visit will be removed from the dataset.
                    This is used to avoid too close visits in the simulated dataset.
                    Default is 1/365 (1 day).

        Returns
        -------
        None
            This method updates the `param_study` attribute of the instance in-place.
        """

        if self.visit_type == VisitType.DATAFRAME:
            patient_number = dict_param["df_visits"].groupby("ID").size().shape[0]

            self.param_study = {
                "patient_number": patient_number,
                "df_visits": dict_param["df_visits"],
            }

        elif self.visit_type == VisitType.RANDOM:
            self.param_study = {
                "patient_number": dict_param["patient_number"],
                "first_visit_mean": dict_param["first_visit_mean"],
                "first_visit_std": dict_param["first_visit_std"],
                "time_follow_up_mean": dict_param["time_follow_up_mean"],
                "time_follow_up_std": dict_param["time_follow_up_std"],
                "distance_visit_mean": dict_param["distance_visit_mean"],
                "distance_visit_std": dict_param["distance_visit_std"],
            }

            # Add optional spacing param if provided
            if "min_spacing_between_visits" in dict_param:
                self.param_study["min_spacing_between_visits"] = dict_param[
                    "min_spacing_between_visits"
                ]

    def _sample_individual_parameters_from_model_parameters(
        self, model: McmcSaemCompatibleModel
    ) -> pd.DataFrame:
        """
        Generate individual parameters for repeated measures simulation, from the model parameters of the loaded model.

        This function samples individual parameters (xi, tau, and source components)
        from a distribution defined by the provided model's parameter.
        Space shifts are computed with the source components and the mixing_matrix.
        It returns the complete set of individual parameters and space shifts in a DataFrame.

        Parameters
        ----------
        model : :class:~.models.abstract_model.McmcSaemCompatibleModel
            A Leaspy model instance containing model parameters,
            among which the mean and standard deviation values for xi, tau, and the mixing matrix.

        Returns
        -------
        pd.DataFrame
            A DataFrame indexed by individual IDs, containing:
            - simulated 'xi' and 'tau': Individual parameters sampled from model distributions.
            - simulated 'sources_X': Latent source components.
            - simulated 'w_X': space shifts derived from the mixing matrix and sources.
        """

        xi_rm = torch.tensor(
            np.random.normal(
                model.hyperparameters["xi_mean"],
                model.parameters["xi_std"],
                self.param_study["patient_number"],
            )
        )

        tau_rm = torch.tensor(
            np.random.normal(
                model.parameters["tau_mean"],
                model.parameters["tau_std"],
                self.param_study["patient_number"],
            )
        )

        if self.visit_type == VisitType.DATAFRAME:
            columns = [str(i) for i in self.param_study["df_visits"]["ID"].unique()]
        else:
            columns = [str(i) for i in range(0, self.param_study["patient_number"])]
        individual_parameters_from_model_parameters = pd.DataFrame(
            [xi_rm, tau_rm],
            index=["xi", "tau"],
            columns=columns,
        ).T

        # Generate the source tensors
        for i in range(model.source_dimension):
            individual_parameters_from_model_parameters[f"sources_{i}"] = torch.tensor(
                np.random.normal(0.0, 1.0, self.param_study["patient_number"]),
                dtype=torch.float32,
            )
            individual_parameters_from_model_parameters[f"sources_{i}"] = (
                individual_parameters_from_model_parameters[f"sources_{i}"]
                - individual_parameters_from_model_parameters[f"sources_{i}"].mean()
            ) / individual_parameters_from_model_parameters[f"sources_{i}"].std()

        patient_source_values_matrix = torch.stack(
            [
                torch.tensor(
                    individual_parameters_from_model_parameters[f"sources_{i}"].values,
                    dtype=torch.float32,
                )
                for i in range(model.source_dimension)
            ],
            dim=1,
        )
        mixing_matrix = model.state.get_tensor_value("mixing_matrix")
        result = torch.matmul(
            mixing_matrix.transpose(0, 1), patient_source_values_matrix.transpose(0, 1)
        )

        space_shifts = pd.DataFrame(
            result.T,
            columns=[f"w_{i}" for i in range(len(self.features))],
            index=individual_parameters_from_model_parameters.index,
        )

        return pd.concat(
            [individual_parameters_from_model_parameters, space_shifts], axis=1
        )

    def _get_leaspy_model(self, model: McmcSaemCompatibleModel) -> None:
        """
        Initialize and store a Leaspy model instance.

        This method creates a new Leaspy object with the 'logistic' model type.
        The resulting instance is stored as an attribute of the class.

        Parameters
        ----------
        model : :class:~.models.abstract_model.McmcSaemCompatibleModel
            A pre-trained Leaspy model to be used for simulation (compute observations).

        Returns
        -------
        None
            This method updates the `leaspy` attribute in-place.
        """

        self._check_logistic_model(model)
        self.model = model

    def _generate_visit_ages(self, df: pd.DataFrame) -> dict:
        """
        Generate visit ages for each individual based on the visit type.

        If the visit type is "dataframe", the visit timepoints are directly extracted
        from the provided DataFrame. Otherwise, synthetic visit ages are generated for
        each individual based on baseline and follow-up ages, with time intervals
        defined by the visit mode  "random"

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame of individual parameters, including 'xi','tau', 'sources' and 'space_shifts'.
            'Tau' is required for generating baseline and follow-up visit ages.

        Returns
        -------
        dict
            Dictionary mapping individual IDs to a list of visit ages (floats).
            - For 'dataframe': uses existing "TIME" values from `df_visits`.
            - For 'random': generates visits with normally-distributed intervals.
        """

        df_ind = df.copy()

        if self.visit_type == VisitType.DATAFRAME:
            return (
                self.param_study["df_visits"]
                .groupby("ID")["TIME"]
                .apply(list)
                .to_dict()
            )

        df_ind["AGE_AT_BASELINE"] = (
            df_ind["tau"].apply(lambda x: x.numpy())
            + pd.DataFrame(
                np.random.normal(
                    self.param_study["first_visit_mean"],
                    self.param_study["first_visit_std"],
                    self.param_study["patient_number"],
                ),
                index=df_ind.index,
            )[0]
        )

        df_ind["AGE_FOLLOW_UP"] = df_ind["AGE_AT_BASELINE"] + np.abs(
            np.random.normal(
                self.param_study["time_follow_up_mean"],
                self.param_study["time_follow_up_std"],
                self.param_study["patient_number"],
            )
        )

        # Generate visit ages for each patients
        dict_timepoints = {}

        for id_ in df_ind.index.values:
            # Get the number of visit per patient
            time = df_ind.loc[id_, "AGE_AT_BASELINE"]
            age_visits = [time]

            while time < df_ind.loc[id_, "AGE_FOLLOW_UP"]:
                if self.visit_type == VisitType.RANDOM:
                    time += np.random.normal(
                        self.param_study["distance_visit_mean"],
                        self.param_study["distance_visit_std"],
                    )

                age_visits.append(time)

            dict_timepoints[id_] = list(age_visits)

        return dict_timepoints

    def _generate_dataset(
        self,
        model: McmcSaemCompatibleModel,
        dict_timepoints: dict,
        individual_parameters_from_model_parameters: pd.DataFrame,
        min_spacing_between_visits: float,
    ) -> pd.DataFrame:
        """
        Generate a simulated dataset based on simulated individual parameters and model timepoints.

        This method simulates observations using estimate function of the Leaspy model. The latter estimates
        values based on the simulated individual parameters: xi, tau and the sources.
        It then adds a beta noise to the simulated values.
        If the visits time are too close to each other, we keep only the first occurrence.
        The user can fix a min delta between two visits. When the delta between the simulated visits is below the threshold, we keep only the first occurrence.
        Min delta spacing is 1 day by default, considering that TIME is in years. If TIME is in another time units, the min_spacing_between_visits will have to be updated

        Parameters
        ----------
        model : :class::~.models.abstract_model.McmcSaemCompatibleModel
            The model used for estimating the individual parameters (in get_ip_rm function) and generating
            the simulated values.

        dict_timepoints : :obj:`dict`
            A dictionary mapping individual IDs to their respective visit timepoints (according to visit_type)

        individual_parameters_from_model_parameters : :obj:`pd.DataFrame`
            DataFrame containing the simulated individual parameters (e.g., 'xi', 'tau', and sources)
            for each individual, used in generating the simulated data.

        min_spacing_between_visits : :obj:`float`, optional
            Default is 1/365 (1 day).
            Minimum delta between visits. This delta has to be in the same unit as the TIME column.
            If two visits are closer than this value, the second visit will be removed from the dataset. This is used to avoid too close visits in the simulated dataset.


        Returns
        -------
        :obj:`pd.DataFrame`
            A DataFrame containing the simulated dataset with ["ID","TIME] as the index
            and features as columns. The dataset includes both the generated values,
            with visits that are too close to each other removed.
        """
        values = self.model.estimate(
            dict_timepoints,
            IndividualParameters().from_dataframe(
                individual_parameters_from_model_parameters[
                    ["xi", "tau"]
                    + [f"sources_{i}" for i in range(model.source_dimension)]
                ]
            ),
        )

        df_long = pd.concat(
            [
                pd.DataFrame(
                    values[id_].clip(max=0.9999999, min=0.00000001),
                    index=pd.MultiIndex.from_product(
                        [[id_], dict_timepoints[id_]], names=["ID", "TIME"]
                    ),
                    columns=[feat + "_no_noise" for feat in self.features],
                )
                for id_ in values.keys()
            ]
        )

        for i, feat in enumerate(self.features):
            if model.parameters["noise_std"].numel() == 1:
                mu = df_long[feat + "_no_noise"]
                var = model.parameters["noise_std"].numpy() ** 2
            else:
                mu = df_long[feat + "_no_noise"]
                var = model.parameters["noise_std"][i].numpy() ** 2

            # Clamp variance where necessary (too big variance and mu too close to 1)
            max_var = mu * (1 - mu)
            adj_var = np.minimum(var, 0.99 * max_var)
            differences = adj_var[adj_var != var]
            for (ID, TIME), adj_val in differences.items():
                warnings.warn(
                    f"Patient {ID} is too advanced in the disease at TIME {np.round(TIME, 3)}. Variance value ({np.round(var, 3)}) out of range for feature {feat}, clamped to {np.round(adj_val, 3)}."
                )

            # Mean and variance parametrization
            alpha_param = mu * ((mu * (1 - mu) / adj_var) - 1)
            beta_param = (1 - mu) * ((mu * (1 - mu) / adj_var) - 1)
            df_long.loc[:, feat] = beta.rvs(alpha_param, beta_param)

        dict_rm_rename = {
            "tau": "RM_TAU",
            "xi": "RM_XI",
            "sources_0": "RM_SOURCES_0",
            "sources_1": "RM_SOURCES_1",
        }

        for i in range(len(self.features)):
            dict_rm_rename[f"w_{i}"] = f"RM_SPACE_SHIFTS_{i}"

        # Put everything in one dataframe
        individual_parameters_from_model_parameters = (
            individual_parameters_from_model_parameters.rename(columns=dict_rm_rename)
        )
        df_sim = df_long[self.features]

        # Drop too close visits
        rounding_options = {
            0: 1,  # 1 year
            1: 0.1,  # 0.1 years ~ 36.5 days
            2: 0.01,  # 0.01 years ~ 3.65 days
            3: 0.001,  # 0.001 years ~ 0.365 days (~1 day) - User will never want precision above 1 day.
        }

        rounding_precision = None
        for precision, val in sorted(rounding_options.items()):
            if val <= min_spacing_between_visits:
                rounding_precision = precision
                break
        df_sim.reset_index(inplace=True)
        df_sim.loc[:, "TIME"] = df_sim["TIME"].round(rounding_precision)
        df_sim.set_index(["ID", "TIME"], inplace=True)
        df_sim = df_sim[~df_sim.index.duplicated()]

        return df_sim



from typing import Type, Union

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.variables.specs import IndividualLatentVariable, PopulationLatentVariable

from .base import (
    AbstractIndividualSampler,
    AbstractPopulationSampler,
    AbstractSampler,
)
from .gibbs import (
    IndividualGibbsSampler,
    PopulationFastGibbsSampler,
    PopulationGibbsSampler,
    PopulationMetropolisHastingsSampler,
)

__all__ = [
    "SamplerFactoryInput",
    "INDIVIDUAL_SAMPLERS",
    "POPULATION_SAMPLERS",
    "sampler_factory",
]

SamplerFactoryInput = Union[str, AbstractSampler]

INDIVIDUAL_SAMPLERS = {"gibbs": IndividualGibbsSampler}

POPULATION_SAMPLERS = {
    "gibbs": PopulationGibbsSampler,
    "fastgibbs": PopulationFastGibbsSampler,
    "metropolis-hastings": PopulationMetropolisHastingsSampler,
}


def sampler_factory(
    sampler: SamplerFactoryInput, variable_type, **kwargs
) -> AbstractSampler:
    """
    Factory for Samplers.

    Parameters
    ----------
    sampler : :class:`.AbstractSampler` or :obj:`str`
        If an instance of a subclass of :class:`.AbstractSampler`, returns the instance.
        If a string, returns a new instance of the appropriate class (with optional parameters `kwargs`).

    variable_type : :class:`.VariableType`
        The type of random variable that the sampler is supposed to sample.

    **kwargs
        Optional parameters for initializing the requested Sampler
        (not used if input is a subclass of :class:`.AbstractSampler`).

    Returns
    -------
    :class:`.AbstractSampler` :
        The desired sampler.

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`:
        If the sampler provided is not supported.
    """
    if isinstance(sampler, AbstractSampler):
        return sampler
    if isinstance(sampler, str):
        kls = _get_sampler_class(sampler, variable_type)
        return kls(**kwargs)
    raise LeaspyAlgoInputError(
        "The provided `sampler` should be a valid instance of `AbstractSampler`, or a string "
        f"among {set(INDIVIDUAL_SAMPLERS).union(POPULATION_SAMPLERS)}."
    )


def _get_sampler_class(sampler_name: str, variable_type):
    if variable_type == IndividualLatentVariable:
        return _get_individual_sampler_class(sampler_name)
    if variable_type == PopulationLatentVariable:
        return _get_population_sampler_class(sampler_name)


def _get_individual_sampler_class(sampler_name: str) -> Type[AbstractIndividualSampler]:
    sampler_name = sampler_name.lower().replace("_", "-")
    kls = INDIVIDUAL_SAMPLERS.get(sampler_name, None)
    if kls is None:
        raise LeaspyAlgoInputError(
            f"Individual sampler '{sampler_name}' is not supported. "
            f"Supported samplers for individual variables are {set(INDIVIDUAL_SAMPLERS)}"
        )
    return kls


def _get_population_sampler_class(sampler_name: str) -> Type[AbstractPopulationSampler]:
    sampler_name = sampler_name.lower().replace("_", "-")
    kls = POPULATION_SAMPLERS.get(sampler_name, None)
    if kls is None:
        raise LeaspyAlgoInputError(
            f"Population sampler '{sampler_name}' is not supported. "
            f"Supported samplers for population variables are {set(POPULATION_SAMPLERS)}"
        )
    return kls



import warnings
from typing import Optional

import numpy as np
import pandas as pd
import torch

from leaspy.exceptions import LeaspyInputError
from leaspy.utils.distributions import discrete_sf_from_pdf
from leaspy.utils.typing import FeatureType, KwargsType

from .data import Data
from .individual_data import IndividualData

__all__ = ["Dataset"]


class Dataset:
    """
    Data container based on :class:`torch.Tensor`, used to run algorithms.

    Parameters
    ----------
    data : :class:`~leaspy.io.data.Data`
        Create `Dataset` from `Data` object
    no_warning : :obj:`bool`, default False
        Whether to deactivate warnings that are emitted by methods of this dataset instance.
        We may want to deactivate them because we rebuild a dataset per individual in scipy minimize.
        Indeed, all relevant warnings certainly occurred for the overall dataset.

    Attributes
    ----------
    headers : :obj:`list` [:obj:`str`]
        Features names
    dimension : :obj:`int`
        Number of features
    n_individuals : :obj:`int`
        Number of individuals
    indices : :obj:`list` [:class:`~leaspy.utils.typing.IDType`]
        Order of patients
    event_time : :obj:`torch.FloatTensor`
        Time of an event, if the event is censored, the time correspond to the last patient observation
    event_bool : :obj:`torch.BoolTensor`
        Boolean to indicate if an event is censored or not: 1 observed, 0 censored
    n_visits_per_individual : :obj:`list` [:obj:`int`]
        Number of visits per individual
    n_visits_max : :obj:`int`
        Maximum number of visits for one individual
    n_visits : :obj:`int`
        Total number of visits
    n_observations_per_ind_per_ft : :obj:`torch.LongTensor`, shape (n_individuals, dimension)
        Number of observations (not taking into account missing values) per individual per feature
    n_observations_per_ft : :obj:`torch.LongTensor`, shape (dimension,)
        Total number of observations per feature
    n_observations : :obj:`int`
        Total number of observations
    timepoints : :obj:`torch.FloatTensor`, shape (n_individuals, n_visits_max)
        Ages of patients at their different visits
    values : :obj:`torch.FloatTensor`, shape (n_individuals, n_visits_max, dimension)
        Values of patients for each visit for each feature
    mask : :obj:`torch.FloatTensor`, shape (n_individuals, n_visits_max, dimension)
        Binary mask associated to values.
        If 1: value is meaningful
        If 0: value is meaningless (either was nan or does not correspond to a real visit - only here for padding)
    L2_norm_per_ft : :obj:`torch.FloatTensor`, shape (dimension,)
        Sum of all non-nan squared values, feature per feature
    L2_norm : scalar :obj:`torch.FloatTensor`
        Sum of all non-nan squared values
    no_warning : :obj:`bool`, default False
        Whether to deactivate warnings that are emitted by methods of this dataset instance.
        We may want to deactivate them because we rebuild a dataset per individual in scipy minimize.
        Indeed, all relevant warnings certainly occurred for the overall dataset.

    _one_hot_encoding : :obj:`dict` [:obj:`bool`, :obj:`torch.LongTensor`]
        Values of patients for each visit for each feature, but tensorized into a one-hot encoding (pdf or sf)
        Shapes of tensors are (n_individuals, n_visits_max, dimension, max_ordinal_level [-1 when `sf=True`])

    Raises
    ------
    :exc:`.LeaspyInputError`
        if data, model or algo are not compatible together.
    """

    def __init__(self, data: Data, *, no_warning: bool = False):
        # Patients information
        self.n_individuals = data.n_individuals
        self.indices = list(data.individuals.keys())

        # Longitudinal outcome information
        self.headers: list[FeatureType] = data.headers
        self.dimension: int = data.dimension
        self.n_visits: int = data.n_visits
        self.timepoints: Optional[torch.FloatTensor] = None
        self.values: Optional[torch.FloatTensor] = None
        self.mask: Optional[torch.FloatTensor] = None
        self.n_observations: Optional[int] = None
        self.n_observations_per_ft: Optional[torch.LongTensor] = None
        self.n_observations_per_ind_per_ft: Optional[torch.LongTensor] = None
        self.n_visits_per_individual: Optional[list[int]] = None
        self.n_visits_max: Optional[int] = None

        # Event information
        self.event_time_name: Optional[str] = data.event_time_name
        self.event_bool_name: Optional[str] = data.event_bool_name
        self.event_time: Optional[torch.FloatTensor] = None
        self.event_bool: Optional[torch.IntTensor] = None

        # Covariate information
        self.covariate_names: Optional[list[str]] = data.covariate_names
        self.covariates: Optional[torch.IntTensor] = None

        # internally used by ordinal models only (cache)
        self._one_hot_encoding: Optional[dict[bool, torch.LongTensor]] = None

        self.L2_norm_per_ft: Optional[torch.FloatTensor] = None
        self.L2_norm: Optional[torch.FloatTensor] = None

        if data.dimension:
            self._construct_values(data)
            self._construct_timepoints(data)
            self._compute_L2_norm()

        if self.event_time_name:
            self._construct_events(data)

        if self.covariate_names:
            self._construct_covariates(data)

        self.no_warning = no_warning

    def _construct_values(self, data: Data):
        """
        Construct the values tensor and the mask tensor from the data.
        The values tensor is of shape (n_individuals, n_visits_max, dimension).

        Parameters
        ----------
        data : :class:`~leaspy.io.data.Data`
            The data from which to construct the values and mask tensors.
        """
        self.n_visits_per_individual = [len(_.timepoints) for _ in data]
        self.n_visits_max = (
            max(self.n_visits_per_individual) if self.n_visits_per_individual else 0
        )  # handle case when empty dataset

        values = torch.zeros((self.n_individuals, self.n_visits_max, self.dimension))
        padding_mask = torch.zeros_like(values)

        # TODO missing values in mask ?

        for i, nb_vis in enumerate(self.n_visits_per_individual):
            # PyTorch 1.10 warns: Creating a tensor from a list of numpy.ndarrays is extremely slow.
            # Please consider converting the list to a single numpy.ndarray with numpy.array() before converting to a tensor.
            # TODO: IndividualData.observations is really badly constructed (list of numpy 1D arrays), we should change this...
            indiv_values = torch.tensor(
                np.array(data[i].observations), dtype=torch.float32
            )
            values[i, 0:nb_vis, :] = indiv_values
            padding_mask[i, 0:nb_vis, :] = 1.0

        mask_missingvalues = (~torch.isnan(values)).float()
        # mask should be 0 on visits outside individual's existing visits (he may have fewer visits than the individual with maximum nb of visits)
        # (we need to enforce it here because we padded values with 0, not with nan, so actual mask is 1 on these fictive values)
        mask = padding_mask * mask_missingvalues

        values[torch.isnan(values)] = 0.0  # Set values of missing values to 0.

        self.values = values
        self.mask = mask

        # number of non-nan observations (different levels of aggregation)
        self.n_observations_per_ind_per_ft = mask.sum(dim=1).int()
        self.n_observations_per_ft = self.n_observations_per_ind_per_ft.sum(dim=0)
        self.n_observations = self.n_observations_per_ft.sum().item()

    def _construct_timepoints(self, data: Data):
        """
        Construct the timepoints tensor from the data.

        Parameters
        ----------
        data : :class:`~leaspy.io.data.Data`
            The data from which to construct the timepoints tensor.
        """
        self.timepoints = torch.zeros((self.n_individuals, self.n_visits_max))
        nbs_vis = [len(_.timepoints) for _ in data]
        for i, nb_vis in enumerate(nbs_vis):
            self.timepoints[i, 0:nb_vis] = torch.tensor(data[i].timepoints)

    def _construct_events(self, data: Data):
        """
        Construct the event time and event boolean tensors from the data.

        Parameters
        ----------
        data : :class:`~leaspy.io.data.Data`
            The data from which to construct the event time and event boolean tensors.
        """
        self.event_time = torch.tensor(
            np.array([_.event_time for _ in data]), dtype=torch.double
        )
        self.event_bool = torch.tensor(
            np.array([_.event_bool for _ in data]), dtype=torch.bool
        )

    def _construct_covariates(self, data: Data):
        """
        Construct the covariates tensor from the data.

        Parameters
        ----------
        data : :class:`~leaspy.io.data.Data`
            The data from which to construct the covariates tensor.
        """
        self.covariates = torch.tensor(
            np.array([_.covariates for _ in data]), dtype=torch.int
        )

    def _compute_L2_norm(self):
        """
        Compute the L2 norm of the values tensor, feature per feature and overall.
        The L2 norm is the sum of the squared values, ignoring nans.
        """
        self.L2_norm_per_ft = torch.sum(
            self.mask.float() * self.values * self.values, dim=(0, 1)
        )  # 1D tensor of shape (dimension,)
        self.L2_norm = self.L2_norm_per_ft.sum()  # sum on all features

    def get_times_patient(self, i: int) -> torch.FloatTensor:
        """
        Get ages for patient number ``i``

        Parameters
        ----------
        i : :obj:`int`
            The index of the patient (<!> not its identifier)

        Returns
        -------
        :obj:`torch.Tensor`, shape (n_obs_of_patient,)
            Contains float
        """
        return self.timepoints[i, : self.n_visits_per_individual[i]]

    def get_event_patient(self, idx_patient: int) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Get ages at event for patient number ``idx_patient``

        Parameters
        ----------
        idx_patient : :obj:`int`
            The index of the patient (<!> not its identifier)

        Returns
        -------
        :obj:`tuple` [:obj:`torch.Tensor`, :obj:`torch.Tensor`] , shape (n_obs_of_patient,)
            Contains float
        """
        if self.event_time is not None and self.event_bool is not None:
            return self.event_time[idx_patient], self.event_bool[idx_patient]
        raise ValueError("Dataset has no event. Please verify your data.")

    def get_covariates_patient(self, idx_patient: int) -> torch.IntTensor:
        """
        Get covariates for patient number ``idx_patient``

        Parameters
        ----------
        idx_patient : :obj:`int`
            The index of the patient (<!> not its identifier)

        Returns
        -------
        :obj:`torch.Tensor`, shape (n_obs_of_patient,)
            Contains float

        Raises
        ------
        :exc:`.ValueError`
            If the dataset has no covariates.
        """
        if self.covariates is not None:
            return self.covariates[idx_patient]
        raise ValueError("Dataset has no covariates. Please verify your data.")

    def get_values_patient(self, i: int, *, adapt_for_model=None) -> torch.FloatTensor:
        """
        Get values for patient number ``i``, with nans.

        Parameters
        ----------
        i : :obj:`int`
            The index of the patient (<!> not its identifier)

        adapt_for_model : None, default or :class:`~leaspy.models.mcmc_saem_compatible.McmcSaemCompatibleModel`
            The values returned are suited for this model.
            In particular:
                * For model with `noise_model='ordinal'` will return one-hot-encoded values [P(X = l), l=0..ordinal_max_level]
                * For model with `noise_model='ordinal_ranking'` will return survival function values [P(X > l), l=0..ordinal_max_level-1]
            If None, we return the raw values, whatever the model is.

        Returns
        -------
        :obj:`torch.Tensor`, shape (n_obs_of_patient, dimension [, extra_dimension_for_ordinal_models])
            Contains float or nans
        """

        # default case (raw values whatever the model)
        values_to_pick_from = self.values
        nans = self.mask[i, : self.n_visits_per_individual[i], :] == 0

        # customization when ordinal model
        if adapt_for_model is not None and getattr(
            adapt_for_model, "is_ordinal", False
        ):
            # we directly fetch the one-hot encoded values (pdf or sf depending on precise `noise_model`)
            values_to_pick_from = self.get_one_hot_encoding(
                sf=adapt_for_model.is_ordinal_ranking,
                ordinal_infos=adapt_for_model.ordinal_infos,
            ).float()

        # we restrict to the right individual and mask the irrelevant data
        values_with_nans = (
            values_to_pick_from[i, : self.n_visits_per_individual[i], ...]
            .clone()
            .detach()
        )
        values_with_nans[nans, ...] = float("nan")

        return values_with_nans

    def to_pandas(self, apply_headers: bool = False) -> pd.DataFrame:
        """
        Convert dataset to a `DataFrame` with ['ID', 'TIME'] index, with all covariates, events and repeated measures if
        apply_headers is False, and only the repeated measures otherwise.

         Parameters
        ----------
        apply_headers : :obj:`bool`
            Enable to select only the columns that are needed for leaspy fit (headers attribute)

        Returns
        -------
        :obj:`pandas.DataFrame`
            DataFrame with index ['ID', 'TIME'] and columns corresponding to the features, events and covariates.

        Raises
        ------
        :exc:`.LeaspyInputError`
            If the index of the DataFrame is not unique or contains invalid values.

        """
        to_concat = []

        for i, idx in enumerate(self.indices):
            ind_pat = IndividualData(idx)

            if self.event_time is not None:
                pat_event_time, pat_event_bool = self.get_event_patient(i)
                ind_pat.add_event(
                    pat_event_time.cpu().tolist(), pat_event_bool.cpu().tolist()
                )

            if self.covariates is not None:
                pat_covariates = self.get_covariates_patient(i)
                ind_pat.add_covariates(pat_covariates.cpu().tolist())

            if self.values is not None:
                times = self.get_times_patient(i).cpu().numpy()
                x = self.get_values_patient(i).cpu().numpy()
                ind_pat.add_observations(times, x)

            to_concat.append(
                ind_pat.to_frame(
                    self.headers,
                    self.event_time_name,
                    self.event_bool_name,
                    self.covariate_names,
                )
            )
        df = pd.concat(to_concat).sort_index()

        if apply_headers:
            df = df[self.headers]
            if not df.index.is_unique:
                raise LeaspyInputError("Index of DataFrame is not unique.")
            if not df.index.to_frame().notnull().all(axis=None):
                raise LeaspyInputError("Index of DataFrame contains invalid values.")
        return df

    def move_to_device(self, device: torch.device) -> None:
        """
        Moves the dataset to the specified device.

        Parameters
        ----------
        device : :obj:`torch.device`
        """
        for attribute_name in dir(self):
            if attribute_name.startswith("__"):
                continue
            attribute = getattr(self, attribute_name)
            if isinstance(attribute, torch.Tensor):
                setattr(self, attribute_name, attribute.to(device))

        ## we have to manually put other variables to the new device

        # Dictionary of one-hot encoded values
        if self._one_hot_encoding is not None:
            self._one_hot_encoding = {
                k: t.to(device) for k, t in self._one_hot_encoding.items()
            }

    def get_one_hot_encoding(
        self, *, sf: bool, ordinal_infos: KwargsType
    ) -> torch.LongTensor:
        """
        Builds the one-hot encoding of ordinal data once and for all and returns it.

        Parameters
        ----------
        sf : :obj:`bool`
            Whether the vector should be the survival function [1(X > l), l=0..max_level-1]
            instead of the probability density function [1(X=l), l=0..max_level]

        ordinal_infos : :class:`~leaspy.utils.typing.KwargsType`
            All the hyperparameters concerning ordinal modelling (in particular maximum level per features)

        Returns
        -------
        :obj:`torch.LongTensor`
            One-hot encoding of data values.

        Raises
        ------
        :exc:`.LeaspyInputError`
            If the values are not non-negative integers or if the features in `ordinal_infos` are not consistent with the dataset headers.
        """
        if self._one_hot_encoding is not None:
            return self._one_hot_encoding[sf]

        ## Check the data & construct the one-hot encodings once for all for fast look-up afterwards

        # Check for values different than non-negative integers
        if (self.values != self.values.round()).any() or (self.values < 0).any():
            raise LeaspyInputError(
                "Please make sure your data contains only integers >= 0 when using ordinal noise modelling."
            )

        # First of all check consistency of features given in ordinal_infos compared to the ones in the dataset (names & order!)
        ordinal_feat_names = list(ordinal_infos["max_levels"])
        if ordinal_feat_names != self.headers:
            raise LeaspyInputError(
                f"Features stored in ordinal model ({ordinal_feat_names}) are not consistent with features in data ({self.headers})"
            )

        # Now check that integers are within the expected range, per feature [0, max_level_ft]
        # (masked values are encoded by 0 at this point)
        vals = self.values.long()
        vals_issues = {
            "unexpected": [],
            "missing": [],
        }
        for ft_i, (ft, max_level_ft) in enumerate(ordinal_infos["max_levels"].items()):
            expected_codes = set(range(0, max_level_ft + 1))  # max level is included

            vals_ft = vals[:, :, ft_i]

            if not self.no_warning:
                # replacing masked values by -1 (which was guaranteed not to be part of input from first check, all >= 0)
                actual_vals_ft = vals_ft.where(
                    self.mask[:, :, ft_i].bool(), torch.tensor(-1)
                )
                actual_codes = set(actual_vals_ft.unique().tolist()).difference({-1})
                unexpected_codes = sorted(actual_codes.difference(expected_codes))
                missing_codes = sorted(expected_codes.difference(actual_codes))
                if len(unexpected_codes):
                    vals_issues["unexpected"].append(
                        f"- {ft} [[0..{max_level_ft}]]: {unexpected_codes} were unexpected"
                    )
                if len(missing_codes):
                    vals_issues["missing"].append(
                        f"- {ft} [[0..{max_level_ft}]]: {missing_codes} are missing"
                    )

            # clip the values (per feature)
            # we must keep this even if no_warning enabled
            vals[:, :, ft_i] = vals_ft.clamp(0, max_level_ft)

        if not self.no_warning and len(vals_issues["unexpected"]):
            warnings.warn(
                f"Some features have unexpected codes (they were clipped to the maximum known level):\n"
                + "\n".join(vals_issues["unexpected"])
            )
        if not self.no_warning and len(vals_issues["missing"]):
            warnings.warn(
                f"Some features have missing codes:\n"
                + "\n".join(vals_issues["missing"])
            )

        # one-hot encode all the values after the checks & clipping
        vals_pdf = torch.nn.functional.one_hot(
            vals, num_classes=ordinal_infos["max_level"] + 1
        )
        # build the survival function by simple (1 - cumsum) and remove the useless P(X >= 0) = 1
        vals_sf = discrete_sf_from_pdf(vals_pdf)
        # cache the values to retrieve them fast afterwards
        self._one_hot_encoding = {False: vals_pdf, True: vals_sf}

        return self._one_hot_encoding[sf]



from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .data import Data
from .dataset import Dataset
from .event_dataframe_data_reader import EventDataframeDataReader
from .factory import (
    DataframeDataReaderFactoryInput,
    DataframeDataReaderNames,
    dataframe_data_reader_factory,
)
from .individual_data import IndividualData
from .joint_dataframe_data_reader import JointDataframeDataReader
from .visit_dataframe_data_reader import VisitDataframeDataReader

__all__ = [
    "AbstractDataframeDataReader",
    "Data",
    "Dataset",
    "EventDataframeDataReader",
    "DataframeDataReaderNames",
    "DataframeDataReaderFactoryInput",
    "dataframe_data_reader_factory",
    "IndividualData",
    "JointDataframeDataReader",
    "VisitDataframeDataReader",
]



"""This module defines the `` class used for all simulation algorithms."""

from abc import abstractmethod

from leaspy.io.data import Dataset
from leaspy.io.data.data import Data
from leaspy.io.outputs import IndividualParameters
from leaspy.io.outputs.result import Result
from leaspy.models import McmcSaemCompatibleModel

from ..base import AlgorithmType, IterativeAlgorithm, ModelType, ReturnType
from ..settings import OutputsSettings

__all__ = ["SimulateAlgorithm"]


class BaseSimulationAlgorithm(IterativeAlgorithm[ModelType, ReturnType]):
    """
    Wrapper class that delegates to the actual implementation of SimulationAlgorithm.
    """

    def set_output_manager(self, output_settings: OutputsSettings) -> None:
        """Set the output manager.

        This is currently not implemented for simulate.
        """
        pass

    def _run(self, model: McmcSaemCompatibleModel) -> Result:
        """Run the simulation pipeline using a leaspy model.

        This method simulates longitudinal data using the given leaspy model.
        It performs the following steps:
        - Retrieves individual parameters (IP) from fixed effects of the model.
        - Loads the specified Leaspy model.
        - Generates visit ages (timepoints) for each individual (based on specifications
        in visits_type from AlgorithmSettings)
        - Simulates observations at those visit ages.
        - Packages the result into a `Result` object, including simulated data,
        individual parameters, and the model's noise standard deviation.

        Parameters
        ----------
        model : :class:~.models.abstract_model.McmcSaemCompatibleModel
            A Leaspy model object previously trained on longitudinal data.

        Returns
        -------
        result_obj : :class:`Result`
            An object containing:
            - `data`: Simulated longitudinal dataset (`Data` object),
            - `individual_parameters`: The individual parameters used in simulation,
            - `noise_std`: Noise standard deviation used in the simulation.
        """

        # Simulate Individual Parameters Repeated Measures
        individual_parameters_from_model_parameters = (
            self._sample_individual_parameters_from_model_parameters(model)
        )

        self._get_leaspy_model(model)

        dict_timepoints = self._generate_visit_ages(
            individual_parameters_from_model_parameters
        )

        min_spacing = self.param_study.get("min_spacing_between_visits", 1 / 365)

        df_sim = self._generate_dataset(
            model,
            dict_timepoints,
            individual_parameters_from_model_parameters,
            min_spacing_between_visits=min_spacing,
        )

        simulated_data = Data.from_dataframe(df_sim)
        result_obj = Result(
            data=simulated_data,
            individual_parameters=individual_parameters_from_model_parameters,
            noise_std=model.parameters["noise_std"].numpy() * 100,
        )
        return result_obj



import re
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch

from leaspy.exceptions import LeaspyConvergenceError
from leaspy.utils.functional import (
    Identity,
    MatMul,
    Mean,
    NamedInputFunction,
    Prod,
    Sqr,
    Std,
    Sum,
    SumDim,
    get_named_parameters,
)
from leaspy.utils.weighted_tensor import WeightedTensor

__all__ = [
    "tensorize_2D",
    "val_to_tensor",
    "serialize_tensor",
    "is_array_like",
    "tensor_to_list",
    "compute_std_from_variance",
    "compute_patient_slopes_distribution",
    "compute_linear_regression_subjects",
    "compute_patient_values_distribution",
    "compute_patient_time_distribution",
    "get_log_velocities",
    "torch_round",
    "compute_ind_param_mean_from_suff_stats_mixture",
    "compute_ind_param_std_from_suff_stats_mixture",
    "compute_ind_param_std_from_suff_stats_mixture_burn_in",
    "compute_probs_from_state",
]


def tensorize_2D(x, unsqueeze_dim: int, dtype=torch.float32) -> torch.Tensor:
    """Convert a scalar or array_like into an, at least 2D, dtype tensor.

    Parameters
    ----------
    x : scalar or array_like
        Element to be tensorized.

    unsqueeze_dim : :obj:`int`
        Dimension to be unsqueezed (0 or -1).
        Meaningful for 1D array-like only (for scalar or vector
        of length 1 it has no matter).

    Returns
    -------
    :class:`torch.Tensor`, at least 2D

    Examples
    --------
    >>> tensorize_2D([1, 2], 0) == tensor([[1, 2]])
    >>> tensorize_2D([1, 2], -1) == tensor([[1], [2])
    """
    # convert to torch.Tensor if not the case
    if not isinstance(x, torch.Tensor):
        x = torch.tensor(x, dtype=dtype)
    # convert dtype if needed
    if x.dtype != dtype:
        x = x.to(dtype)
    # if tensor is less than 2-dimensional add dimensions
    while x.dim() < 2:
        x = x.unsqueeze(dim=unsqueeze_dim)
    # postcondition: x.dim() >= 2
    return x


def val_to_tensor(val, shape: Optional[tuple] = None):
    if not isinstance(val, (torch.Tensor, WeightedTensor)):
        val = torch.tensor(val)
    if shape is not None:
        val = val.view(shape)  # no expansion here
    return val


def serialize_tensor(v, *, indent: str = "", sub_indent: str = "") -> str:
    """Nice serialization of floats, torch tensors (or numpy arrays)."""
    from torch._tensor_str import PRINT_OPTS as torch_print_opts

    if isinstance(v, (str, bool, int)):
        return str(v)
    if isinstance(v, np.ndarray):
        return str(v.tolist())
    if isinstance(v, float) or getattr(v, "ndim", -1) == 0:
        # for 0D tensors / arrays the default behavior is to print all digits...
        # change this!
        return f"{v:.{1 + torch_print_opts.precision}g}"
    if isinstance(v, (list, frozenset, set, tuple)):
        try:
            return serialize_tensor(
                torch.tensor(list(v)), indent=indent, sub_indent=sub_indent
            )
        except Exception:
            return str(v)
    if isinstance(v, dict):
        if not len(v):
            return ""
        subs = [
            f"{p} : "
            + serialize_tensor(vp, indent="  ", sub_indent=" " * len(f"{p} : ["))
            for p, vp in v.items()
        ]
        lines = [indent + _ for _ in "\n".join(subs).split("\n")]
        return "\n" + "\n".join(lines)
    # torch.tensor, np.array, ...
    # in particular you may use `torch.set_printoptions` and `np.set_printoptions` globally
    # to tune the number of decimals when printing tensors / arrays
    v_repr = str(v)
    # remove tensor prefix & possible device/size/dtype suffixes
    v_repr = re.sub(r"^[^\(]+\(", "", v_repr)
    v_repr = re.sub(r"(?:, device=.+)?(?:, size=.+)?(?:, dtype=.+)?\)$", "", v_repr)
    # adjust justification
    return re.sub(r"\n[ ]+([^ ])", rf"\n{sub_indent}\1", v_repr)


def is_array_like(v: Any) -> bool:
    try:
        len(v)  # exclude np.array(scalar) or torch.tensor(scalar)
        return hasattr(v, "__getitem__")  # exclude set
    except Exception:
        return False


def tensorize_2D(x, unsqueeze_dim: int, dtype=torch.float32) -> torch.Tensor:
    """Convert a scalar or array_like into an, at least 2D, dtype tensor.

    Parameters
    ----------
    x : scalar or array_like
        Element to be tensorized.

    unsqueeze_dim : :obj:`int`
        Dimension to be unsqueezed (0 or -1).
        Meaningful for 1D array-like only (for scalar or vector
        of length 1 it has no matter).

    Returns
    -------
    :class:`torch.Tensor`, at least 2D

    Examples
    --------
    >>> tensorize_2D([1, 2], 0) == tensor([[1, 2]])
    >>> tensorize_2D([1, 2], -1) == tensor([[1], [2])
    """
    # convert to torch.Tensor if not the case
    if not isinstance(x, torch.Tensor):
        x = torch.tensor(x, dtype=dtype)
    # convert dtype if needed
    if x.dtype != dtype:
        x = x.to(dtype)
    # if tensor is less than 2-dimensional add dimensions
    while x.dim() < 2:
        x = x.unsqueeze(dim=unsqueeze_dim)
    # postcondition: x.dim() >= 2
    return x


def val_to_tensor(val, shape: Optional[tuple] = None):
    if not isinstance(val, (torch.Tensor, WeightedTensor)):
        val = torch.tensor(val)
    if shape is not None:
        val = val.view(shape)  # no expansion here
    return val


def serialize_tensor(v, *, indent: str = "", sub_indent: str = "") -> str:
    """Nice serialization of floats, torch tensors (or numpy arrays)."""
    from torch._tensor_str import PRINT_OPTS as torch_print_opts

    if isinstance(v, (str, bool, int)):
        return str(v)
    if isinstance(v, np.ndarray):
        return str(v.tolist())
    if isinstance(v, float) or getattr(v, "ndim", -1) == 0:
        # for 0D tensors / arrays the default behavior is to print all digits...
        # change this!
        return f"{v:.{1 + torch_print_opts.precision}g}"
    if isinstance(v, (list, frozenset, set, tuple)):
        try:
            return serialize_tensor(
                torch.tensor(list(v)), indent=indent, sub_indent=sub_indent
            )
        except Exception:
            return str(v)
    if isinstance(v, dict):
        if not len(v):
            return ""
        subs = [
            f"{p} : "
            + serialize_tensor(vp, indent="  ", sub_indent=" " * len(f"{p} : ["))
            for p, vp in v.items()
        ]
        lines = [indent + _ for _ in "\n".join(subs).split("\n")]
        return "\n" + "\n".join(lines)
    # torch.tensor, np.array, ...
    # in particular you may use `torch.set_printoptions` and `np.set_printoptions` globally
    # to tune the number of decimals when printing tensors / arrays
    v_repr = str(v)
    # remove tensor prefix & possible device/size/dtype suffixes
    v_repr = re.sub(r"^[^\(]+\(", "", v_repr)
    v_repr = re.sub(r"(?:, device=.+)?(?:, size=.+)?(?:, dtype=.+)?\)$", "", v_repr)
    # adjust justification
    return re.sub(r"\n[ ]+([^ ])", rf"\n{sub_indent}\1", v_repr)


def is_array_like(v: Any) -> bool:
    try:
        len(v)  # exclude np.array(scalar) or torch.tensor(scalar)
        return hasattr(v, "__getitem__")  # exclude set
    except Exception:
        return False


def tensor_to_list(x: Union[list, torch.Tensor]) -> list:
    """
    Convert input tensor to list.

    Parameters
    ----------
    x : :obj:`list` or :obj:`torch.Tensor`
        Input tensor or list to be converted.

    Returns
    -------
    :obj:`list`
        List converted from tensor input, or original list if input was not a tensor.

    Raises
    ------
    :exc:`NotImplementedError`
        If the input is a `WeightedTensor`, as this functionality is not yet implemented.
    """
    if isinstance(x, (np.ndarray, torch.Tensor)):
        return x.tolist()
    if isinstance(x, WeightedTensor):
        raise NotImplementedError("TODO")
    return x


def compute_std_from_variance(
    variance: torch.Tensor,
    varname: str,
    tol: float = 1e-5,
) -> torch.Tensor:
    """
    Check that variance is strictly positive and return its square root, otherwise fail with a convergence error.
    If variance is multivariate check that all components are strictly positive.
    TODO? a full Bayesian setting with good priors on all variables should prevent such convergence issues.

    Parameters
    ----------
    variance : :obj:`torch.Tensor`
        The variance we would like to convert to a std-dev.
    varname : :obj:`str`
        The name of the variable.
    tol : :obj:`float`, optional
        The lower bound on variance, under which the converge error is raised.
        Default=1e-5.

    Returns
    -------
    :obj: `torch.Tensor` :
        The standard deviation from the variance.

    Raises
    ------
    :exc:`.LeaspyConvergenceError`
        If the variance is less than the specified tolerance, indicating a convergence issue.
    """

    if (variance < tol).any():
        raise LeaspyConvergenceError(
            f"The parameter '{varname}' collapsed to zero, which indicates a convergence issue.\n"
            "Start by investigating what happened in the logs of your calibration and try to double check:"
            "\n- your training dataset (not enough subjects and/or visits? too much missing data?)"
            "\n- the hyperparameters of your Leaspy model (`source_dimension` too low or too high? "
            "observation model not suited to your data?)"
            "\n- the hyperparameters of your calibration algorithm"
        )

    return variance.sqrt()


def compute_ind_param_std_from_suff_stats(
    state: Dict[str, torch.Tensor],
    ip_values: torch.Tensor,
    ip_sqr_values: torch.Tensor,
    *,
    ip_name: str,
    dim: int,
    **kws,
):
    """
    Maximization rule, from the sufficient statistics, of the standard-deviation
    of Gaussian prior for individual latent variables.

    Parameters
    ----------
    state : Dict[str, torch.Tensor]
    ip_values : torch.Tensor
    ip_sqr_values : torch.Tensor
    ip_name : str
    dim : int
    """
    ip_old_mean = state[f"{ip_name}_mean"]
    ip_cur_mean = torch.mean(ip_values, dim=dim)
    ip_var_update = torch.mean(ip_sqr_values, dim=dim) - 2 * ip_old_mean * ip_cur_mean
    ip_var = ip_var_update + ip_old_mean**2
    return compute_std_from_variance(ip_var, varname=f"{ip_name}_std", **kws)


def compute_ind_param_mean_from_suff_stats_mixture(
    state: Dict[str, torch.Tensor],
    *,
    ip_name: str,
) -> torch.Tensor:
    ind_var = state[f"{ip_name}"]
    nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
    nll_cluster = -nll_regul_ind_sum_ind

    probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.0))

    if ip_name == "sources":  # special treatment due to the extra dimension
        ind_var_expanded = ind_var.unsqueeze(-1)
        probs_expanded = probs_ind.unsqueeze(1)
        result = ind_var_expanded * probs_expanded
    else:
        result = probs_ind * ind_var

    result = result.sum(dim=0) / probs_ind.sum(dim=0)

    return result


def compute_ind_param_std_from_suff_stats_mixture(
    state: Dict[str, torch.Tensor],
    ip_values: torch.Tensor,
    ip_sqr_values: torch.Tensor,
    *,
    ip_name: str,
    dim: int,
    **kws,
):
    ip_old_mean = state[f"{ip_name}_mean"]
    ip_cur_mean = torch.mean(ip_values, dim=0)
    ip_var_update = torch.mean(ip_sqr_values, dim=0) - 2 * ip_old_mean * ip_cur_mean
    ip_var = ip_var_update + ip_old_mean**2
    std = ip_var.sqrt()

    nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
    nll_cluster = -nll_regul_ind_sum_ind

    probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.0))

    result = (probs_ind * std).sum(dim=0) / probs_ind.sum(dim=0)

    return result


def compute_ind_param_std_from_suff_stats_mixture_burn_in(
    state: Dict[str, torch.Tensor],
    *,
    ip_name: str,
) -> torch.Tensor:
    ind_var = state[f"{ip_name}"].std(dim=0)
    nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
    nll_cluster = -nll_regul_ind_sum_ind

    probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.0))

    result = (probs_ind * ind_var).sum(dim=0) / probs_ind.sum(dim=0)

    return result


def compute_probs_from_state(
    state: Dict[str, torch.Tensor],
) -> torch.Tensor:
    nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
    n_inds = nll_regul_ind_sum_ind.shape[0]
    nll_cluster = -nll_regul_ind_sum_ind
    probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.0))

    return probs_ind.sum(dim=0) / n_inds


def compute_patient_slopes_distribution(
    df: pd.DataFrame,
    *,
    max_inds: Optional[int] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute linear regression slopes and their standard deviations for each feature.

    Parameters
    ----------
    df : :obj:`pd.DataFrame`
        DataFrame containing individual scores.
    max_inds : :obj:`int`, optional
        Restrict computation to first `max_inds` individuals.
        Default="None"

    Returns
    -------
    :obj:`Tuple`[:obj:`torch.Tensor`, :obj:`torch.Tensor`]:
        Tuple with :
        - [0] : torch.Tensor of shape (n_features,) - Regression slopes
        - [1] : torch.Tensor of shape (n_features,) - Standard deviation of the slopes
    """
    d_regress_params = compute_linear_regression_subjects(df, max_inds=max_inds)
    slopes_mu, slopes_sigma = [], []

    for ft, df_regress_ft in d_regress_params.items():
        slopes_mu.append(df_regress_ft["slope"].mean())
        slopes_sigma.append(df_regress_ft["slope"].std())

    return torch.tensor(slopes_mu), torch.tensor(slopes_sigma)


def compute_linear_regression_subjects(
    df: pd.DataFrame,
    *,
    max_inds: Optional[int] = None,
) -> Dict[str, pd.DataFrame]:
    """
    Linear Regression on each feature to get intercept & slopes

    Parameters
    ----------
    df : :obj:`pd.DataFrame`
        Contains the individual scores.
    max_inds : :obj:`int`, optional
        Restrict computation to first `max_inds` individuals.
        Default="None"

    Returns
    -------
    :obj: `Dict`[:obj:`str`, :obj:`pd.DataFrame`]:
        Dictionary with :
        - keys : feature names
        - values : DataFrame with :
            - index : Individual IDs
            - columns : 'intercept', 'slope'

    """
    regression_parameters = {}

    for feature, data in df.items():
        data = data.dropna()
        n_visits = data.groupby("ID").size()
        indices_train = n_visits[n_visits >= 2].index
        if max_inds:
            indices_train = indices_train[:max_inds]
        data_train = data.loc[indices_train]
        regression_parameters[feature] = (
            data_train.groupby("ID").apply(_linear_regression_against_time).unstack(-1)
        )

    return regression_parameters


def _linear_regression_against_time(data: pd.Series) -> Dict[str, float]:
    """
    Return intercept & slope of a linear regression of series values
    against time (present in series index).

    Parameters
    ----------
    data : :obj:`pd.Series`
        Series with time index and values to regress

    Returns
    -------
    :obj: `Dict`[:obj:`str`, :obj: `float`]:
        Dictionary with:
        - keys : 'intercept', 'slope'
        - values : intercept & slope of the linear regression
    """
    from scipy.stats import linregress

    y = data.values
    t = data.index.get_level_values("TIME").values
    slope, intercept, r_value, p_value, std_err = linregress(t, y)
    return {"intercept": intercept, "slope": slope}


def compute_patient_values_distribution(
    df: pd.DataFrame,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Returns means and standard deviations for the features of the given dataset values.

    Parameters
    ----------
    df : :obj:`pd.DataFrame`
        Contains the individual scores.

    Returns
    -------
    :obj: Tuple[:obj:`torch.Tensor`, :obj:`torch.Tensor`]:
        Tuple with:
        - [0] : torch.Tensor of shape (n_features,) - Means of the features
        - [1] : torch.Tensor of shape (n_features,) - Standard deviations of the features
    """
    return torch.tensor(df.mean().values), torch.tensor(df.std().values)


def compute_patient_time_distribution(
    df: pd.DataFrame,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Returns mu / sigma of given dataset times.

    Parameters
    ----------
    df : :obj:`pd.DataFrame`
        Contains the individual scores.

    Returns
    -------
    :obj:`Tuple`[:obj:`torch.Tensor`, :obj:`torch.Tensor`]:
        Tuple with:
        - [0] : torch.Tensor - Mean of the times
        - [1] : torch.Tensor - Standard deviation of the times
    """
    times = df.index.get_level_values("TIME").values
    return torch.tensor([times.mean()]), torch.tensor([times.std()])


def get_log_velocities(
    velocities: torch.Tensor, features: List[str], *, min: float = 1e-2
) -> torch.Tensor:
    """
    Get the log of the velocities, clamping them to `min` if negative.

    Parameters
    ----------
    velocities : :obj:`torch.Tensor`
        The velocities to be clamped and logged.
    features : :obj:`List`[:obj:`str`]
        The names of the features corresponding to the velocities.
    min : :obj:`float`, optional
        The minimum value to clamp the velocities to.
        Default=1e-2

    Returns
    -------
    :obj:`torch.Tensor` :
        The log of the clamped velocities.

    Raises
    ------
    :obj:`Warning`
        If some negative velocities are provided.
    """

    neg_velocities = velocities <= 0
    if neg_velocities.any():
        warnings.warn(
            f"Mean slope of individual linear regressions made at initialization is negative for "
            f"{[f for f, vel in zip(features, velocities) if vel <= 0]}: not properly handled in model..."
        )
    return velocities.clamp(min=min).log()


def torch_round(t: torch.FloatTensor, *, tol: float = 1 << 16) -> torch.FloatTensor:
    """
    Multiplies the tensor by `tol`, applies standard rounding, then scales back.
    This effectively rounds values to the nearest multiple of `1.0 / tol`.

    Parameters
    ----------
    t : :obj:`torch.FloatTensor`
        The tensor to be rounded.

    tol : :obj:`float`, optional
        The tolerance factor controlling rounding precision (higher = finer rounding).
        Default=1 << 16 (65536).
        This corresponds to rounding to ~ 10**-4.8.

    Returns
    -------
    :obj:`torch.FloatTensor` :
        The rounded tensor with the same shape as input `t`.
    """
    return (t * tol).round() * (1.0 / tol)



from enum import Enum
from pathlib import Path
from typing import Union

import pandas as pd

from leaspy.io.outputs import IndividualParameters
from leaspy.models import BaseModel

__all__ = [
    "DatasetName",
    "load_dataset",
    "load_individual_parameters",
    "load_model",
    "get_dataset_path",
    "get_individual_parameter_path",
    "get_model_path",
]


class DatasetName(str, Enum):
    """
    Enum for the names of the datasets available in Leaspy.
    The names correspond to the files in the `data` folder.
    """

    ALZHEIMER = "alzheimer"
    PARKINSON = "parkinson"
    PARKINSON_PUTAMEN = "parkinson-putamen"
    PARKINSON_PUTAMEN_TRAIN_TEST = "parkinson-putamen-train_and_test"


def get_dataset_path(name: Union[str, DatasetName]) -> Path:
    """
    Get the path to the dataset file.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.datasets.loader.DatasetName`
        The name of the dataset.

    Returns
    -------
    :obj:`pathlib.Path`
        The path to the dataset file.

    Example:
    --------
    >>> from leaspy.datasets.loader import get_dataset_path
    >>> path = get_dataset_path("alzheimer")
    """
    name = DatasetName(name)
    current_folder = Path(__file__).parent.resolve()
    return current_folder / "data" / f"{name.value}.csv"


def get_individual_parameter_path(name: Union[str, DatasetName]) -> Path:
    """
    Get the path to the individual parameters file.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.datasets.loader.DatasetName`
        The name of the dataset.

    Returns
    -------
    :obj:`pathlib.Path`
        The path to the individual parameters file.

    Raises
    ------
    :exc:`ValueError`
        If the dataset does not have individual parameters, such as `parkinson-putamen-train_and_test`.
    """
    name = DatasetName(name)
    if name == DatasetName.PARKINSON_PUTAMEN_TRAIN_TEST:
        raise ValueError(
            f"No individual parameter sample for the dataset {name.value}."
        )
    current_folder = Path(__file__).parent.resolve()
    return (
        current_folder
        / "individual_parameters"
        / f"{name.value}-individual_parameters.csv"
    )


def get_model_path(name: Union[str, DatasetName]) -> Path:
    """
    Get the path to the model parameters file.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.datasets.loader.DatasetName`
        The name of the dataset.

    Returns
    -------
    :obj:`pathlib.Path`
        The path to the model parameters file.

    Raises
    ------
    :exc:`ValueError`
        If the dataset does not have a model, such as `parkinson-putamen-train_and_test`.
    """
    name = DatasetName(name)
    if name == DatasetName.PARKINSON_PUTAMEN_TRAIN_TEST:
        raise ValueError(f"No model instance for the dataset {name.value}.")
    current_folder = Path(__file__).parent.resolve()
    return current_folder / "model_parameters" / f"{name.value}-model_parameters.json"


def load_dataset(dataset_name: Union[str, DatasetName]) -> pd.DataFrame:
    """
    Load synthetic longitudinal observations mimicking cohort of subjects with neurodegenerative disorders.

    Parameters
    ----------
    dataset_name : :obj:`str` or :class:`DatasetName`
        The name of the dataset to load.

    Returns
    -------
    :obj:`pandas.DataFrame`
        The DataFrame containing the IDs, timepoints and observations.

    Notes
    -----
    All `DataFrames` have the same structures.

    * Index: a :class:`pandas.MultiIndex` - ``['ID', 'TIME']`` which contain IDs and timepoints.
        The `DataFrame` is sorted by index. So, one line corresponds to one visit for one subject.
        The `DataFrame` having `'train_and_test'` in their name also have ``'SPLIT'`` as the third
        index level. It differentiate `train` and `test` data.

    * Columns: One column correspond to one feature (or score).
    """
    df = pd.read_csv(get_dataset_path(dataset_name), dtype={"ID": str})
    if "SPLIT" in df.columns:
        df.set_index(["ID", "TIME", "SPLIT"], inplace=True)
    else:
        df.set_index(["ID", "TIME"], inplace=True)
    return df.sort_index()


def load_individual_parameters(name: Union[str, DatasetName]) -> IndividualParameters:
    """
    Load a Leaspy instance with a model already calibrated on the synthetic dataset corresponding to the name
    of the instance.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.datasets.loader.DatasetName`
        The name of the individual parameters to load.

    Returns
    -------
    :class:`~leaspy.io.outputs.IndividualParameters`
        Leaspy instance with a model already calibrated.
    """
    return IndividualParameters.load(str(get_individual_parameter_path(name)))


def load_model(name: Union[str, DatasetName]) -> BaseModel:
    """Load a model already calibrated on the synthetic dataset corresponding to the name of the instance.

    Parameters
    ----------
    name : :obj:`str` or :class:`~leaspy.datasets.loader.DatasetName`
        The name of the instance to load.

    Returns
    -------
    :class:`~leaspy.models.BaseModel`
        Model instance already calibrated.
    """
    return BaseModel.load(str(get_model_path(name)))



from operator import itemgetter

import pandas as pd
import torch

import leaspy
from leaspy.exceptions import LeaspyInputError, LeaspyModelInputError

XI_STD = 0.5
TAU_STD = 5.0
NOISE_STD = 0.1
SOURCES_STD = 1.0


def initialize_parameters(model, dataset, method="default") -> tuple:
    """
    Initialize the model's group parameters given its name & the scores of all subjects.

    Under-the-hood it calls an initialization function dedicated for the `model`:
        * :func:`.initialize_linear` (including when `univariate`)
        * :func:`.initialize_logistic` (including when `univariate`)
        * :func:`.initialize_logistic_parallel`

    It is automatically called during :meth:`.Leaspy.fit`.

    Parameters
    ----------
    model : :class:`.AbstractModel`
        The model to initialize.
    dataset : :class:`.Dataset`
        Contains the individual scores.
    method : str
        Must be one of:
            * ``'default'``: initialize at mean.
            * ``'random'``:  initialize with a gaussian realization with same mean and variance.

    Returns
    -------
    parameters : dict [str, :class:`torch.Tensor`]
        Contains the initialized model's group parameters.

    Raises
    ------
    :exc:`.LeaspyInputError`
        If no initialization method is known for model type / method
    """

    # we convert once for all dataset to pandas dataframe for convenience
    df = dataset.to_pandas().dropna(how="all").sort_index()
    assert df.index.is_unique
    assert df.index.to_frame().notnull().all(axis=None)
    if model.features != df.columns.tolist():
        raise LeaspyInputError(
            f"Features mismatch between model and dataset: {model.features} != {df.columns}"
        )

    if method == "lme":
        raise NotImplementedError("legacy")
        # return lme_init(model, df) # support kwargs?


def get_lme_results(
    df: pd.DataFrame, n_jobs=-1, *, with_random_slope_age=True, **lme_fit_kwargs
):
    r"""
    Fit a LME on univariate (per feature) time-series (feature vs. patients' ages with varying intercept & slope)

    Parameters
    ----------
    df : :class:`pd.DataFrame`
        Contains all the data (with nans)
    n_jobs : int
        Number of jobs in parallel when multiple features to init
        Not used for now, buggy
    with_random_slope_age : bool (default True)
        Has LME model a random slope per age (otherwise only a random intercept).
    **lme_fit_kwargs
        Kwargs passed to 'lme_fit' (such as `force_independent_random_effects`, default True)

    Returns
    -------
    dict
        {param: str -> param_values_for_ft: torch.Tensor(nb_fts, \*shape_param)}
    """

    # defaults for LME Fit algorithm settings
    lme_fit_kwargs = {"force_independent_random_effects": True, **lme_fit_kwargs}

    # @delayed
    def fit_one_ft(df_ft):
        data_ft = leaspy.Data.from_dataframe(df_ft)
        lsp_lme_ft = leaspy.Leaspy("lme", with_random_slope_age=with_random_slope_age)
        algo = leaspy.AlgorithmSettings("lme_fit", **lme_fit_kwargs)  # seed=seed

        lsp_lme_ft.fit(data_ft, algo)

        return lsp_lme_ft.model.parameters

    # res = Parallel(n_jobs=n_jobs)(delayed(fit_one_ft)(s.dropna().to_frame()) for ft, s in df.items())
    res = list(fit_one_ft(s.dropna().to_frame()) for ft, s in df.items())

    # output a dict of tensor stacked by feature, indexed by param
    param_names = next(iter(res)).keys()

    return {
        param_name: torch.stack([torch.tensor(res_ft[param_name]) for res_ft in res])
        for param_name in param_names
    }


def lme_init(model, df: pd.DataFrame, fact_std=1.0, **kwargs):
    """
    Initialize the model's group parameters.

    Parameters
    ----------
    model : :class:`.AbstractModel`
        The model to initialize (must be an univariate or multivariate linear or logistic manifold model).
    df : :class:`pd.DataFrame`
        Contains the individual scores (with nans).
    fact_std : float
        Multiplicative factor to apply on std-dev (tau, xi, noise) found naively with LME
    **kwargs
        Additional kwargs passed to :func:`.get_lme_results`

    Returns
    -------
    parameters : dict [str, `torch.Tensor`]
        Contains the initialized model's group parameters.

    Raises
    ------
    :exc:`.LeaspyInputError`
        If model is not supported for this initialization
    """
    raise NotImplementedError("OLD")

    name = model.name
    noise_model = model.noise_model  # has to be set directly at model init and not in algo settings step to be available here

    if not isinstance(noise_model, AbstractGaussianNoiseModel):
        raise LeaspyModelInputError(
            f"`lme` initialization is only compatible with Gaussian noise models, not {noise_model}."
        )

    multiv = "univariate" not in name

    # print('Initialization with linear mixed-effects model...')
    lme = get_lme_results(df, **kwargs)
    # print()

    # init
    params = {}

    v0_lin = (lme["fe_params"][:, 1] / lme["ages_std"]).clamp(min=1e-2)  # > exp(-4.6)

    if "linear" in name:
        # global tau mean (arithmetic mean of ages mean)
        params["tau_mean"] = lme["ages_mean"].mean()

        params["g"] = lme["fe_params"][:, 0] + v0_lin * (
            params["tau_mean"] - lme["ages_mean"]
        )
        params["v0" if multiv else "xi_mean"] = v0_lin.log()

    # elif name in ['logistic_parallel']:
    #    # deltas = torch.zeros((model.dimension - 1,)) ?
    #    pass # TODO...
    elif name in ["logistic", "univariate_logistic"]:
        """
        # global tau mean (arithmetic mean of inflexion point per feature)
        t0_ft = lme['ages_mean'] + (.5 - lme['fe_params'][:, 0]) / v0_lin # inflexion pt
        params['tau_mean'] = t0_ft.mean()
        """

        # global tau mean (arithmetic mean of ages mean)
        params["tau_mean"] = lme["ages_mean"].mean()

        # positions at this tau mean
        pos_ft = lme["fe_params"][:, 0] + v0_lin * (
            params["tau_mean"] - lme["ages_mean"]
        )

        # parameters under leaspy logistic formulation
        g = 1 / pos_ft.clamp(min=1e-2, max=1 - 1e-2) - 1
        params["g"] = g.log()  # -4.6 ... 4.6

        v0 = g / (1 + g) ** 2 * 4 * v0_lin  # slope from lme at inflexion point

        # if name == 'logistic_parallel':
        #    # a common speed for all features!
        #    params['xi_mean'] = v0.log().mean() # or log of fts-mean?
        # else:
        params["v0" if multiv else "xi_mean"] = v0.log()

    else:
        raise LeaspyInputError(
            f"Model '{name}' is not supported in `lme` initialization."
        )

    ## Dispersion of individual parameters
    # approx. dispersion on tau (-> on inflexion point when logistic)
    tau_var_ft = lme["cov_re"][:, 0, 0] / v0_lin**2
    params["tau_std"] = (
        fact_std * (1 / tau_var_ft).mean() ** -0.5
    )  # harmonic mean on variances per ft

    # approx dispersion on alpha and then xi
    alpha_var_ft = lme["cov_re"][:, 1, 1] / lme["fe_params"][:, 1] ** 2
    xi_var_ft = (
        1 / 2 + (1 / 4 + alpha_var_ft) ** 0.5
    ).log()  # because alpha = exp(xi) so var(alpha) = exp(2*var_xi) - exp(var_xi)
    params["xi_std"] = fact_std * (1 / xi_var_ft).mean() ** -0.5

    # Residual gaussian noise
    if isinstance(noise_model, GaussianScalarNoiseModel):
        # arithmetic mean on variances
        params["noise_std"] = (
            fact_std * (lme["noise_std"] ** 2).mean().reshape((1,)) ** 0.5
        )  # 1D tensor
    else:
        # one noise-std per feature
        params["noise_std"] = fact_std * lme["noise_std"]

    # For multivariate models, xi_mean == 0.
    if name in ["linear", "logistic"]:  # isinstance(model, MultivariateModel)
        params["xi_mean"] = torch.tensor(0.0)

    if multiv:  # including logistic_parallel
        params["betas"] = torch.zeros((model.dimension - 1, model.source_dimension))
        params["sources_mean"] = torch.tensor(0.0)
        params["sources_std"] = torch.tensor(SOURCES_STD)

    return params



from typing import Tuple

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import DictParamsTorch

from .abstract_attributes import AbstractAttributes

__all__ = ["AbstractManifoldModelAttributes"]


class AbstractManifoldModelAttributes(AbstractAttributes):
    """
    Abstract base class for attributes of leaspy manifold models.

    Contains the common attributes & methods of the different attributes classes.
    Such classes are used to update the models' attributes.

    Parameters
    ----------
    name : str
    dimension : int
    source_dimension : int (default None)

    Attributes
    ----------
    name : str (default None)
        Name of the associated leaspy model.
    dimension : int
    source_dimension : int
    univariate : bool
        Whether model is univariate or not (i.e. dimension == 1)
    has_sources : bool
        Whether model has sources or not (not univariate and source_dimension >= 1)
    update_possibilities : set[str]
        Contains the available parameters to update. Different models have different parameters.

    positions : :class:`torch.Tensor` [dimension] (default None)
        <!> Depending on the model it does not correspond to the same thing.
    velocities : :class:`torch.Tensor` [dimension] (default None)
        Vector of velocities for each feature (positive components).
        For multivariate models only (except for parallel model as it is useless).
    orthonormal_basis : :class:`torch.Tensor` [dimension, dimension - 1] (default None)
        For multivariate and multivariate parallel models, with source_dimension >= 1.
    betas : :class:`torch.Tensor` [dimension - 1, source_dimension] (default None)
        For multivariate and multivariate parallel models, with source_dimension >= 1.
    mixing_matrix : :class:`torch.Tensor` [dimension, source_dimension] (default None)
        Matrix A such that w_i = A * s_i.
        For multivariate and multivariate parallel models, with source_dimension >= 1.

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        if any inconsistent parameter.
    """

    def __init__(self, name: str, dimension: int, source_dimension: int):
        super().__init__(name, dimension, source_dimension)

        self.positions: torch.FloatTensor = torch.FloatTensor()
        self.velocities: torch.FloatTensor = torch.FloatTensor()
        self.mixing_matrix: torch.FloatTensor = torch.FloatTensor()
        self.betas: torch.FloatTensor = torch.FloatTensor()
        self.orthonormal_basis = None
        self.update_possibilities = {"all", "g", "v0", "v0_collinear"}

        if self.has_sources:
            self.update_possibilities.add("betas")

    def get_attributes(
        self,
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:
        """
        Returns the attributes of the model, which is a tuple of three torch tensors.

        For the precise definitions of those attributes please refer to the exact
        attributes class associated to your model.

        Returns
        -------
        positions: `torch.Tensor`
        velocities: `torch.Tensor`
        mixing_matrix: `torch.Tensor`
        """
        return self.positions, self.velocities, self.mixing_matrix

    def _compute_velocities(self, values: DictParamsTorch):
        """
        Update the attribute ``velocities`` (only for multivariate - not parallel - models).

        Parameters
        ----------
        values : dict [str, `torch.Tensor`]
        """
        self.velocities = torch.exp(values["v0"])

    def _compute_betas(self, values: DictParamsTorch):
        """
        Update the attribute ``betas``.

        Parameters
        ----------
        values : dict [str, `torch.Tensor`]
        """
        self.betas = values["betas"].clone()

    def _compute_Q(
        self,
        dgamma_t0: torch.FloatTensor,
        G_metric: torch.FloatTensor,
        strip_col: int = 0,
    ):
        """
        Householder decomposition, adapted for a non-Euclidean inner product defined by:
        (1) :math:`< x, y >Metric(p) = < x, G(p) y >Eucl = xT G(p) y`, where:
        :math:`G(p)` is the symmetric positive-definite (SPD) matrix defining the metric at point `p`.

        The Euclidean case is the special case where `G` is the identity matrix.
        Product-metric is a special case where `G(p)` is a diagonal matrix (identified to a vector)
        whose components are all > 0.

        It is used in child classes to compute and set in-place the ``orthonormal_basis`` attribute
        given the time-derivative of the geodesic at initial time and the `G_metric`.
        The first component of the full orthonormal basis is a vector collinear `G_metric x dgamma_t0` that we get rid of.

        The orthonormal basis we construct is always orthonormal for the Euclidean canonical inner product.
        But all (but first) vectors of it lie in the sub-space orthogonal (for canonical inner product) to `G_metric * dgamma_t0`
        which is the same thing that being orthogonal to `dgamma_t0` for the inner product implied by the metric.

        [We could do otherwise if we'd like a full orthonormal basis, w.r.t. the non-Euclidean inner product.
        But it'd imply to compute G^(-1/2) & G^(1/2) which may be computationally costly in case we don't have direct access to them
        (for the special case of product-metric it is easy - just the component-wise inverse (sqrt'ed) of diagonal)
        TODO are there any advantages/drawbacks of one method over the other except this one?
        TODO are there any biases between features when only considering Euclidean orthonormal basis?]

        Parameters
        ----------
        dgamma_t0 : :class:`torch.FloatTensor` 1D
            Time-derivative of the geodesic at initial time.
            It may also be a vector collinear to it without any change to the result.

        G_metric : scalar, `torch.FloatTensor` 0D, 1D or 2D-square
            The `G(p)` defining the metric as referred in equation (1) just before :
                * If 0D (scalar): `G` is proportional to the identity matrix
                * If 1D (vector): `G` is a diagonal matrix (diagonal components > 0)
                * If 2D (square matrix): `G` is general (SPD)

        strip_col : int in 0..model_dimension-1 (default 0)
            Which column of the basis should be the one collinear to `dgamma_t0` (that we get rid of)

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            if incoherent metric `G_metric`
        """

        # enforce `G_metric` to be a tensor
        if not isinstance(G_metric, torch.Tensor):
            G_metric = torch.tensor(G_metric)  # convert from scalar...

        # compute the vector that others columns should be orthogonal to, w.r.t canonical inner product
        G_shape = G_metric.shape
        if len(G_shape) == 0:  # 0D
            if G_metric.item() <= 0:
                raise LeaspyModelInputError("Incoherent negative scalar metric.")

            dgamma_t0 = G_metric.item() * dgamma_t0  # homothetic
        elif len(G_shape) == 1:  # 1D
            if not (G_metric > 0).all():
                raise LeaspyModelInputError(
                    "Incoherent 1D metric with negative values."
                )
            if len(G_metric) != self.dimension:
                raise LeaspyModelInputError(
                    f"Incoherent 1D metric size: {len(G_metric)} != {self.dimension}."
                )

            dgamma_t0 = G_metric * dgamma_t0  # component-wise multiplication of vectors
        elif len(G_shape) == 2:  # matrix (general case)
            if len(G_metric) != self.dimension:
                raise LeaspyModelInputError(
                    f"Incoherent 2D metric shape: {G_shape} != {(self.dimension, self.dimension)}."
                )

            dgamma_t0 = G_metric @ dgamma_t0  # matrix multiplication
        else:
            raise LeaspyModelInputError(
                "Unexpected metric of dim > 2 when computing orthonormal basis."
            )

        """
        Automatically choose the best column to strip?
        <!> Not a good idea because it could fluctuate over iterations making mixing_matrix unstable!
            (betas should slowly readapt to the permutation...)
        #strip_col = dgamma_t0.abs().argmax().item()
        #strip_col = v_metric_normalization.argmin().item()
        """

        assert 0 <= strip_col < self.dimension
        ej = torch.zeros(self.dimension)
        ej[strip_col] = 1.0

        alpha = -torch.sign(dgamma_t0[strip_col]) * torch.norm(dgamma_t0)
        u_vector = dgamma_t0 - alpha * ej
        v_vector = u_vector / torch.norm(u_vector)

        ## Classical Householder method (to get an orthonormal basis for the canonical inner product)
        ## Q = I_n - 2 v • v'
        q_matrix = torch.eye(self.dimension) - 2 * v_vector.view(-1, 1) * v_vector

        # first component of basis is a unit vector (for metric norm) collinear to `dgamma_t0`
        # self.orthonormal_basis = q_matrix[:, 1:]

        # concat columns (get rid of the one collinear to `dgamma_t0`)
        self.orthonormal_basis = torch.cat(
            (q_matrix[:, :strip_col], q_matrix[:, strip_col + 1 :]), dim=1
        )

    @staticmethod
    def _mixing_matrix_utils(
        linear_combination_values: torch.FloatTensor, matrix: torch.FloatTensor
    ) -> torch.FloatTensor:
        """
        Intermediate function used to test the good behaviour of the class' methods.

        Parameters
        ----------
        linear_combination_values : :class:`torch.FloatTensor`
        matrix : :class:`torch.FloatTensor`

        Returns
        -------
        :class:`torch.FloatTensor`
        """
        return torch.mm(matrix, linear_combination_values)

    @staticmethod
    def _check_collinearity_vectors(*vectors: torch.FloatTensor) -> bool:
        """
        Returns True if all vectors are collinear [or all zeros].

        This may be a useful helper function to avoid unneeded orthonormal basis re-computations.
        (Not used for now - only "declaration" of sure collinearity to speed-up)

        Precondition on vectors: 0D or 1D torch float tensors of same shapes
        """
        linalg_mod = getattr(torch, "linalg")
        torch_matrix_rank = getattr(linalg_mod, "matrix_rank", torch.matrix_rank)
        return (
            torch_matrix_rank(torch.stack(vectors).view(len(vectors), -1)).item() <= 1
        )

    def _compute_mixing_matrix(self):
        """
        Update the attribute ``mixing_matrix``.
        """
        self.mixing_matrix = self._mixing_matrix_utils(
            self.betas, self.orthonormal_basis
        )



"""Defines the noise model factory."""

from enum import Enum
from typing import Dict, Type, Union

from leaspy.exceptions import LeaspyModelInputError

from ._base import ObservationModel
from ._bernoulli import BernoulliObservationModel
from ._gaussian import FullGaussianObservationModel
from ._weibull import (
    WeibullRightCensoredObservationModel,
    WeibullRightCensoredWithSourcesObservationModel,
)

__all__ = [
    "ObservationModelNames",
    "ObservationModelFactoryInput",
    "OBSERVATION_MODELS",
    "observation_model_factory",
]


class ObservationModelNames(Enum):
    """Enumeration defining the possible names for observation models."""

    GAUSSIAN_DIAGONAL = "gaussian-diagonal"
    GAUSSIAN_SCALAR = "gaussian-scalar"
    BERNOULLI = "bernoulli"
    WEIBULL_RIGHT_CENSORED = "weibull-right-censored"
    WEIBULL_RIGHT_CENSORED_WITH_SOURCES = "weibull-right-censored-with-sources"

    @classmethod
    def from_string(cls, model_name: str):
        try:
            return cls(model_name.lower().replace("_", "-"))
        except ValueError:
            raise NotImplementedError(
                f"The requested ObservationModel {model_name} is not implemented. "
                f"Valid observation model names are: {[elt.value for elt in cls]}."
            )


ObservationModelFactoryInput = Union[str, ObservationModelNames, ObservationModel]

OBSERVATION_MODELS: Dict[ObservationModelNames, Type[ObservationModel]] = {
    ObservationModelNames.GAUSSIAN_DIAGONAL: FullGaussianObservationModel,
    ObservationModelNames.GAUSSIAN_SCALAR: FullGaussianObservationModel,
    ObservationModelNames.BERNOULLI: BernoulliObservationModel,
    ObservationModelNames.WEIBULL_RIGHT_CENSORED: WeibullRightCensoredObservationModel,
    ObservationModelNames.WEIBULL_RIGHT_CENSORED_WITH_SOURCES: WeibullRightCensoredWithSourcesObservationModel,
}


def observation_model_factory(
    model: ObservationModelFactoryInput, **kwargs
) -> ObservationModel:
    """
    Factory for observation models.

    Parameters
    ----------
    model : :obj:`str` or :class:`.ObservationModel` or :obj:`dict` [ :obj:`str`, ...]
        - If an instance of a subclass of :class:`.ObservationModel`, returns the instance.
        - If a string, then returns a new instance of the appropriate class (with optional parameters `kws`).
        - If a dictionary, it must contain the 'name' key and other initialization parameters.
    **kwargs
        Optional parameters for initializing the requested observation model when a string.

    Returns
    -------
    :class:`.ObservationModel` :
        The desired observation model.

    Raises
    ------
    :exc:`.LeaspyModelInputError` :
        If `model` is not supported.
    """
    dimension = kwargs.pop("dimension", None)
    n_clusters = kwargs.pop("n_clusters", None)
    if isinstance(model, ObservationModel):
        return model
    if isinstance(model, str):
        model = ObservationModelNames.from_string(model)
    if isinstance(model, ObservationModelNames):
        if model == ObservationModelNames.GAUSSIAN_DIAGONAL:
            if dimension is None:
                raise NotImplementedError(
                    "WIP: dimension / features should be provided to "
                    f"init the obs_model = {ObservationModelNames.GAUSSIAN_DIAGONAL}."
                )
            return FullGaussianObservationModel.with_noise_std_as_model_parameter(
                dimension
            )
        if model == ObservationModelNames.GAUSSIAN_SCALAR:
            return FullGaussianObservationModel.with_noise_std_as_model_parameter(1)
        if model == ObservationModelNames.WEIBULL_RIGHT_CENSORED:
            return WeibullRightCensoredObservationModel.default_init(kwargs=kwargs)
        if model == ObservationModelNames.WEIBULL_RIGHT_CENSORED_WITH_SOURCES:
            return WeibullRightCensoredWithSourcesObservationModel.default_init(
                kwargs=kwargs
            )
        return OBSERVATION_MODELS[model](**kwargs)
    raise LeaspyModelInputError(
        "The provided `model` should be a valid instance of `ObservationModel`, "
        f"or a string among {[c.value for c in ObservationModelNames]}."
        f"Instead, {model} of type {type(model)} was provided."
    )



"""This module defines the `ScipyMinimize` class."""

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from pprint import pformat
from typing import Type

import numpy as np
import torch
from joblib import Parallel, delayed
from scipy.optimize import minimize

from leaspy.io.data import Data, Dataset
from leaspy.io.outputs.individual_parameters import IndividualParameters
from leaspy.models import McmcSaemCompatibleModel
from leaspy.utils.typing import DictParamsTorch
from leaspy.variables.specs import (
    IndividualLatentVariable,
    LatentVariable,
    VariableName,
)
from leaspy.variables.state import State

from ..base import AlgorithmName
from ..settings import AlgorithmSettings
from .base import PersonalizeAlgorithm

__all__ = ["ScipyMinimizeAlgorithm"]


@dataclass(frozen=True)
class _AffineScaling:
    """
    Affine scaling used for individual latent variables, so that gradients
    are of the same order of magnitude in scipy minimize.
    """

    loc: torch.Tensor
    scale: torch.Tensor

    @property
    def shape(self) -> tuple[int, ...]:
        shape = self.loc.shape
        assert self.scale.shape == shape
        return shape

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @classmethod
    def from_latent_variable(cls, var: LatentVariable, state: State) -> _AffineScaling:
        """Natural scaling for latent variable: (mode, stddev)."""

        mode = var.prior.mode.call(state)
        stddev = var.prior.stddev.call(state)

        # Ensure they're 1D tensors
        mode = mode.reshape(1) if mode.ndim == 0 else mode
        stddev = stddev.reshape(1) if stddev.ndim == 0 else stddev

        return cls(mode, stddev)


@dataclass
class _AffineScalings1D:
    """
    Util class to deal with scaled 1D tensors, that are concatenated
    together in a single 1D tensor (in order).
    """

    scalings: dict[VariableName, _AffineScaling]
    slices: dict[VariableName, slice] = field(init=False, repr=False, compare=False)
    length: int = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        assert all(
            scl.ndim == 1 for scl in self.scalings.values()
        ), "Individual latent variables should all be 1D vectors"
        dims = {n: scl.shape[0] for n, scl in self.scalings.items()}
        import operator
        from itertools import accumulate

        cumdims = (0,) + tuple(accumulate(dims.values(), operator.add))
        slices = {n: slice(cumdims[i], cumdims[i + 1]) for i, n in enumerate(dims)}
        object.__setattr__(self, "slices", slices)
        object.__setattr__(self, "length", cumdims[-1])

    def __len__(self) -> int:
        return self.length

    def stack(self, x: dict[VariableName, torch.Tensor]) -> torch.Tensor:
        """
        Stack the provided mapping in a multidimensional numpy array.

        Parameters
        ----------
        x : Dict[VarName, torch.Tensor]
            Mapping to stack.

        Return
        ------
        np.ndarray :
            Stacked array of values.
        """
        return torch.cat([x[n].float() for n, _ in self.scalings.items()])

    def unstack(self, x: torch.Tensor) -> dict[VariableName, torch.Tensor]:
        """ "
        Unstack the provided concatenated array.

        Parameters
        ----------
        x : np.ndarray
            Concatenated array to unstack.

        Return
        ------
        dict :
            Mapping from variable names to their tensor values.
        """
        return {n: x[None, self.slices[n]].float() for n, _ in self.scalings.items()}

    def unscaling(self, x: np.ndarray) -> dict[VariableName, torch.Tensor]:
        """
        Unstack the concatenated array and unscale
        each element to bring it back to its natural scale.

        Parameters
        ----------
        x : np.ndarray
            The concatenated array to scale.

        Returns
        -------
        dict :
            Mapping from variable name to tensor values scaled.
        """
        x_unscaled = torch.cat(
            [
                # unsqueeze 1 dimension at left
                scaling.loc + scaling.scale * x[self.slices[n]]
                for n, scaling in self.scalings.items()
            ]
        )
        return self.unstack(x_unscaled)

    def scaling(self, x: dict[VariableName, torch.Tensor]) -> np.ndarray:
        """
        Scale and concatenate provided mapping of values
        from their natural scale to the defined scale.

        Parameters
        ----------
        x : :obj:`dict`[VarName, torch.Tensor]
            The mapping to unscale.

        Return
        ------
        np.ndarray :
            Concatenated array of scaled values.
        """
        x_stacked = self.stack(x)
        return (
            torch.cat(
                [
                    # unsqueeze 1 dimension at left
                    (x_stacked[self.slices[n]].float() - scaling.loc) / scaling.scale
                    for n, scaling in self.scalings.items()
                ]
            )
            .detach()
            .numpy()
        )

    @classmethod
    def from_state(
        cls, state: State, var_type: Type[LatentVariable]
    ) -> _AffineScalings1D:
        """
        Get the affine scalings of latent variables so their gradients have the same
        order of magnitude during optimization.
        """
        return cls(
            {
                var_name: _AffineScaling.from_latent_variable(var, state)
                for var_name, var in state.dag.sorted_variables_by_type[
                    var_type
                ].items()
            }
        )


class ScipyMinimizeAlgorithm(
    PersonalizeAlgorithm[McmcSaemCompatibleModel, IndividualParameters]
):
    """Gradient descent based algorithm to compute individual parameters, `i.e.` personalizing a model for a given set of subjects.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        Settings for the algorithm, including the `custom_scipy_minimize_params`
        parameter, which contains keyword arguments passed to
        :func:`scipy.optimize.minimize`.

    Attributes
    ----------
    scipy_minimize_params : :obj:`dict`
        Keyword arguments for :func:`scipy.optimize.minimize`, with default values
        depending on the usage of a jacobian (cf.
        `ScipyMinimize.DEFAULT_SCIPY_MINIMIZE_PARAMS_WITH_JACOBIAN` and
        `ScipyMinimize.DEFAULT_SCIPY_MINIMIZE_PARAMS_WITHOUT_JACOBIAN`).
        Customization is possible via the `custom_scipy_minimize_params` in
        :class:`.AlgorithmSettings`.

    format_convergence_issues : :obj:`str`
       A format string for displaying convergence issues, which can use the
        following variables:
            - `patient_id`: :obj:`str`
            - `optimization_result_pformat`: :obj:`str`
            - `optimization_result_obj`: dict-like
        The default format is defined in
        `ScipyMinimize.DEFAULT_FORMAT_CONVERGENCE_ISSUES`, but it can be
        customized via the `custom_format_convergence_issues` parameter.

    logger : None or callable :obj:`str` -> None
        The function used to display convergence issues returned by :func:`scipy.optimize.minimize`.
        By default, convergence issues are printed only if the BFGS optimization method is not used.
        This can  be customized by setting the `logger` attribute in :class:`.AlgorithmSettings`.
    """

    name: AlgorithmName = AlgorithmName.PERSONALIZE_SCIPY_MINIMIZE
    DEFAULT_SCIPY_MINIMIZE_PARAMS_WITH_JACOBIAN = {
        "method": "BFGS",
        "options": {
            "gtol": 1e-2,
            "maxiter": 200,
        },
    }
    DEFAULT_SCIPY_MINIMIZE_PARAMS_WITHOUT_JACOBIAN = {
        "method": "Powell",
        "options": {
            "xtol": 1e-4,
            "ftol": 1e-4,
            "maxiter": 200,
        },
    }
    DEFAULT_FORMAT_CONVERGENCE_ISSUES = (
        "<!> {patient_id}:\n{optimization_result_pformat}"
    )
    regularity_factor: float = 1.0

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.scipy_minimize_params = self.algo_parameters.get(
            "custom_scipy_minimize_params", None
        )
        if self.scipy_minimize_params is None:
            if self.algo_parameters["use_jacobian"]:
                self.scipy_minimize_params = (
                    self.DEFAULT_SCIPY_MINIMIZE_PARAMS_WITH_JACOBIAN
                )
            else:
                self.scipy_minimize_params = (
                    self.DEFAULT_SCIPY_MINIMIZE_PARAMS_WITHOUT_JACOBIAN
                )
        self.format_convergence_issues = self.algo_parameters.get(
            "custom_format_convergence_issues", None
        )
        if self.format_convergence_issues is None:
            self.format_convergence_issues = self.DEFAULT_FORMAT_CONVERGENCE_ISSUES

        # use a sentinel object to be able to set a custom logger=None
        _sentinel = object()
        self.logger = getattr(settings, "logger", _sentinel)
        if self.logger is _sentinel:
            self.logger = self._default_logger

    def _default_logger(self, msg: str) -> None:
        # we dynamically retrieve the method of `scipy_minimize_params` so that if we requested jacobian
        # but had to fall back to without jacobian we do print messages!
        if not self.scipy_minimize_params.get("method", "BFGS").upper() == "BFGS":
            print("\n" + msg + "\n")

    def _get_normalized_grad_tensor_from_grad_dict(
        self,
        dict_grad_tensors: DictParamsTorch,
        model: McmcSaemCompatibleModel,
    ):
        """
        From a dict of gradient tensors per param (without normalization),
        returns the full tensor of gradients (= for all params, consecutively):
            * concatenated with conventional order of x0
            * normalized because we derive w.r.t. "standardized" parameter (adimensional gradient)
        """
        raise NotImplementedError("TODO...")
        to_cat = [
            dict_grad_tensors["xi"] * model.parameters["xi_std"],
            dict_grad_tensors["tau"] * model.parameters["tau_std"],
        ]
        if "univariate" not in model.name and model.source_dimension > 0:
            to_cat.append(
                dict_grad_tensors["sources"] * model.parameters["sources_std"]
            )

        return torch.cat(to_cat, dim=-1)  # 1 individual at a time

    def _get_regularity(
        self,
        model: McmcSaemCompatibleModel,
        individual_parameters: DictParamsTorch,
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """
        Compute the regularity term (and its gradient) of a patient given his individual parameters for a given model.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            Model used to compute the group average parameters.

        individual_parameters : :obj:`dict`[:obj:`str`, :class:`torch.Tensor` [n_ind,n_dims_param]]
            Individual parameters as a dict

        Returns
        -------
        regularity : :class:`torch.Tensor` [n_individuals]
            Regularity of the patient(s) corresponding to the given individual parameters.
            (Sum on all parameters)

        regularity_grads : :obj:`dict`[param_name: :obj:`str`, :class:`torch.Tensor` [n_individuals, n_dims_param]]
            Gradient of regularity term with respect to individual parameters.
        """
        d_regularity, d_regularity_grads = (
            model.compute_regularity_individual_parameters(individual_parameters)
        )
        tot_regularity = sum(d_regularity.values(), 0.0)

        return tot_regularity, d_regularity_grads

    def obj_no_jac(
        self, x: np.ndarray, state: State, scaling: _AffineScalings1D
    ) -> float:
        """
        Objective loss function to minimize in order to get patient's individual parameters.

        Parameters
        ----------
        x : numpy.ndarray
            Individual standardized parameters
            At initialization x is full of zeros (mode of priors, scaled by std-dev)
        state : :class:`.State`
            The cloned model state that is dedicated to the current individual.
            In particular, individual data variables for the current individual are already loaded into it.
        scaling : _AffineScalings1D
            The scaling to be used for individual latent variables.

        Returns
        -------
        objective : :obj:`float`
            Value of the loss function (negative log-likelihood).
        """

        ips = scaling.unscaling(x)
        for ip, ip_val in ips.items():
            state[ip] = ip_val
        loss = state["nll_attach"] + self.regularity_factor * state["nll_regul_ind_sum"]
        return loss.item()

    def obj_with_jac(
        self, x: np.ndarray, state: State, scaling: _AffineScalings1D
    ) -> tuple[float, torch.Tensor]:
        """
        Objective loss function to minimize in order to get patient's individual parameters,
        together with its jacobian w.r.t to each of `x` dimension.

        Parameters
        ----------
        x : numpy.ndarray
            Individual standardized parameters
            At initialization x is full of zeros (mode of priors, scaled by std-dev)
        state : :class:`.State`
            The cloned model state that is dedicated to the current individual.
            In particular, individual data variables for the current individual are already loaded into it.
        scaling : _AffineScalings1D
            The scaling to be used for individual latent variables.

        Returns
        -------
        2-tuple (as expected by :func:`scipy.optimize.minimize` when ``jac=True``)
            * objective : :obj:`float`
            * gradient : array-like[float] with same length as `x` (= all dimensions of individual latent variables, concatenated)
        """
        raise NotImplementedError("TODO...")

        individual_parameters = self._pull_individual_parameters(x, model)
        predictions = model.compute_individual_tensorized(
            dataset.timepoints, individual_parameters
        )

        nll_regul, d_nll_regul_grads = self._get_regularity(
            model, individual_parameters
        )
        nll_attach = model.noise_model.compute_nll(
            dataset, predictions, with_gradient=with_gradient
        )
        if with_gradient:
            nll_attach, nll_attach_grads_fact = nll_attach

        # we must sum separately the terms due to implicit broadcasting
        nll = nll_attach.squeeze(0).sum() + nll_regul.squeeze(0)

        if not with_gradient:
            return nll.item()

        nll_regul_grads = self._get_normalized_grad_tensor_from_grad_dict(
            d_nll_regul_grads, model
        ).squeeze(0)

        d_preds_grads = model.compute_jacobian_tensorized(
            dataset.timepoints, individual_parameters
        )
        # put derivatives consecutively in the right order
        # --> output shape [1, n_tpts, n_fts [, n_ordinal_lvls], n_dims_params]
        preds_grads = self._get_normalized_grad_tensor_from_grad_dict(
            d_preds_grads, model
        ).squeeze(0)

        grad_dims_to_sum = tuple(range(0, preds_grads.ndim - 1))
        nll_attach_grads = (
            preds_grads * nll_attach_grads_fact.squeeze(0).unsqueeze(-1)
        ).sum(dim=grad_dims_to_sum)

        nll_grads = nll_attach_grads + nll_regul_grads

        return nll.item(), nll_grads

    def _get_individual_parameters_patient(
        self,
        state: State,
        *,
        scaling: _AffineScalings1D,
        with_jac: bool,
        patient_id: str,
    ) -> tuple[dict[str, torch.Tensor], torch.Tensor]:
        """
        Compute the individual parameter by minimizing the objective loss function with scipy solver.

        Parameters
        ----------
        state : :class:`.State`
            The cloned model state that is dedicated to the current individual.
            In particular, data variables for the current individual are already loaded into it.
        scaling : _AffineScalings1D
            The scaling to be used for individual latent variables.
        with_jac : :obj:`bool`
            Should we speed-up the minimization by sending exact gradient of optimized function?
        patient_id : :obj:`str`
            ID of patient (essentially here for logging purposes when no convergence)

        Returns
        -------
        pyt_individual_params : :obj:`dict`[:obj:`str`, :class:`torch.Tensor` [1,n_dims_param]]
            Individual parameters as a dict of tensors.
        reconstruction_loss : :class:`torch.Tensor`
            Model canonical loss (content & shape depend on noise model).
            TODO
        """
        obj = self.obj_with_jac if with_jac else self.obj_no_jac

        # Initialize the optimization at the state's values
        # for individual parameters
        initial_point = {
            n: state.get_tensor_value(n)[0] for n in state.dag.individual_variable_names
        }

        res = minimize(
            obj,
            jac=with_jac,
            x0=scaling.scaling(initial_point),
            args=(state, scaling),
            **self.scipy_minimize_params,
        )
        pyt_individual_params = scaling.unscaling(res.x)
        # TODO/WIP: we may want to return residuals MAE or RMSE instead (since nll is not very interpretable...)
        # loss = model.compute_canonical_loss_tensorized(patient_dataset, pyt_individual_params)
        loss = self.obj_no_jac(res.x, state, scaling)

        if not res.success and self.logger:
            # log full results if optimization failed
            # including mean of reconstruction loss for this subject on all his
            # personalization visits, but per feature
            res["reconstruction_loss"] = loss
            res["individual_parameters"] = pyt_individual_params

            cvg_issue = self.format_convergence_issues.format(
                patient_id=patient_id,
                optimization_result_obj=res,
                optimization_result_pformat=pformat(res, indent=1),
            )
            self.logger(cvg_issue)

        return pyt_individual_params, loss

    def _get_individual_parameters_patient_master(
        self,
        state: State,
        *,
        scaling: _AffineScalings1D,
        progress: tuple[int, int],
        with_jac: bool,
        patient_id: str,
    ):
        """
        Compute individual parameters of all patients given a leaspy model & a leaspy dataset.

        Parameters
        ----------
        state : :class:`.State`
            The cloned model state that is dedicated to the current individual.
            In particular, individual data variables for the current individual are already loaded into it.
        scaling : _AffineScalings1D
            The scaling to be used for individual latent variables.
        progress : tuple[int >= 0, int > 0]
            Current progress in loop (n, out-of-N).
        with_jac : :obj:`bool`
            Should we speed-up the minimization by sending exact gradient of optimized function?
            Should we speed-up the minimization by sending exact gradient of optimized function?
        patient_id : :obj:`str`
            ID of patient (essentially here for logging purposes when no convergence)

        Returns
        -------
        :class:`.IndividualParameters`
            Contains the individual parameters of all patients.
        """
        individual_params_tensorized, _ = self._get_individual_parameters_patient(
            state, scaling=scaling, with_jac=with_jac, patient_id=patient_id
        )

        if self.algo_parameters.get("progress_bar", True):
            self._display_progress_bar(*progress, suffix="subjects")

        # TODO/WIP: change this really dirty stuff (hardcoded...)
        # transformation is needed because of current `IndividualParameters` expectations... --> change them
        return {
            k: v.detach().squeeze(0).tolist()
            for k, v in individual_params_tensorized.items()
        }

    def is_jacobian_implemented(self, model: McmcSaemCompatibleModel) -> bool:
        """Check that the jacobian of model is implemented."""
        # TODO/WIP: quick hack for now
        return any("jacobian" in var_name for var_name in model.dag)
        # default_individual_params = self._pull_individual_parameters(self._initialize_parameters(model), model)
        # empty_tpts = torch.tensor([[]], dtype=torch.float32)
        # try:
        #    model.compute_jacobian_tensorized(empty_tpts, default_individual_params)
        #    return True
        # except NotImplementedError:
        #    return False

    def _compute_individual_parameters(
        self, model: McmcSaemCompatibleModel, dataset: Dataset, **kwargs
    ) -> IndividualParameters:
        """
        Compute individual parameters of all patients given a leaspy model & a leaspy dataset.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            Model used to compute the group average parameters.
        dataset : :class:`.Dataset` class object
            Contains the individual scores.

        Returns
        -------
        :class:`.IndividualParameters`
            Contains the individual parameters of all patients.
        """
        # Easier to pass a Dataset with 1 individual rather than individual times, values
        # to avoid duplicating code in noise model especially
        df = dataset.to_pandas()
        import pandas as pd

        assert pd.api.types.is_string_dtype(
            df.index.dtypes["ID"]
        ), "Individuals ID should be strings"

        if "joint" in model.name:
            data_type = "joint"
            factory_kws = {"nb_events": model.nb_events}
        else:
            data_type = "visit"
            factory_kws = {}

        data = Data.from_dataframe(
            df,
            drop_full_nan=False,
            warn_empty_column=False,
            data_type=data_type,
            factory_kws=factory_kws,
        )

        datasets = {
            idx: Dataset(data[[idx]], no_warning=True) for idx in dataset.indices
        }

        # Fetch model internal state (latent pop. vars should be OK)
        state = model.state
        # Fixed scalings for individual parameters
        ips_scalings = _AffineScalings1D.from_state(
            state, var_type=IndividualLatentVariable
        )

        # Clone model states (1 per individual with the appropriate dataset loaded into each of them)
        states = {}
        for idx in dataset.indices:
            states[idx] = state.clone(disable_auto_fork=True)
            model.put_data_variables(states[idx], datasets[idx])
            # Get an individual initial value for minimisation
            model.put_individual_parameters(states[idx], datasets[idx])

        if self.algo_parameters.get("progress_bar", True):
            self._display_progress_bar(-1, dataset.n_individuals, suffix="subjects")

        # optimize by sending exact gradient of optimized function?
        with_jac = self.algo_parameters["use_jacobian"]
        if with_jac and not self.is_jacobian_implemented(model):
            warnings.warn(
                "In `scipy_minimize` you requested `use_jacobian=True` but it "
                f"is not implemented in your model {model.name}. "
                "Falling back to `use_jacobian=False`..."
            )
            with_jac = False
            if self.algo_parameters.get("custom_scipy_minimize_params", None) is None:
                # reset default `scipy_minimize_params`
                self.scipy_minimize_params = (
                    self.DEFAULT_SCIPY_MINIMIZE_PARAMS_WITHOUT_JACOBIAN
                )
            # TODO? change default logger as well?

        ind_p_all = Parallel(n_jobs=self.algo_parameters["n_jobs"])(
            delayed(self._get_individual_parameters_patient_master)(
                state_pat,
                scaling=ips_scalings,
                progress=(it_pat, dataset.n_individuals),
                with_jac=with_jac,
                patient_id=id_pat,
            )
            # TODO use Parallel + tqdm instead of custom progress bar...
            for it_pat, (id_pat, state_pat) in enumerate(states.items())
        )

        individual_parameters = IndividualParameters()
        for id_pat, ind_params_pat in zip(dataset.indices, ind_p_all):
            individual_parameters.add_individual_parameters(str(id_pat), ind_params_pat)

        return individual_parameters



"""This module defines the `AbstractPersonalizeAlgo` class used for all personalize algorithms."""

from abc import abstractmethod

from leaspy.io.data import Dataset
from leaspy.io.outputs import IndividualParameters

from ..base import AlgorithmType, IterativeAlgorithm, ModelType, ReturnType
from ..settings import OutputsSettings

__all__ = ["PersonalizeAlgorithm"]


class PersonalizeAlgorithm(IterativeAlgorithm[ModelType, ReturnType]):
    """Abstract class for `personalize` algorithm.

    Estimation of individual parameters of a given `Data` file with
    a frozen model (already estimated, or loaded from known parameters).

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        Settings of the algorithm.

    Attributes
    ----------
    name : :obj:`str`
        Algorithm's name.
    seed : :obj:`int`, optional
        Algorithm's seed (default None).
    algo_parameters : :obj:`dict`
        Algorithm's parameters.

    See Also
    --------
    :meth:`.Leaspy.personalize`
    """

    family: AlgorithmType = AlgorithmType.PERSONALIZE

    def set_output_manager(self, output_settings: OutputsSettings) -> None:
        """Set the output manager.

        This is currently not implemented for personalize.
        """
        pass

    def _run(self, model: ModelType, dataset: Dataset, **kwargs) -> ReturnType:
        r"""Main personalize function, wraps the abstract :meth:`._get_individual_parameters` method.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            A subclass object of leaspy `McmcSaemCompatibleModel`.

        dataset : :class:`.Dataset`
            Dataset object build with leaspy class objects Data, algo & model

        Returns
        -------
        individual_parameters : :class:`.IndividualParameters`
            Contains individual parameters.
        """
        return self._compute_individual_parameters(model, dataset, **kwargs)

    @abstractmethod
    def _compute_individual_parameters(
        self, model: ModelType, dataset: Dataset, **kwargs
    ) -> IndividualParameters:
        raise NotImplementedError()



"""`ObservationModel` defines the common interface for observation models in Leaspy."""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Optional,
)
from typing import (
    Mapping as TMapping,
)

from leaspy.io.data.dataset import Dataset
from leaspy.utils.functional import SumDim
from leaspy.utils.weighted_tensor import WeightedTensor, sum_dim
from leaspy.variables.distributions import SymbolicDistribution
from leaspy.variables.specs import (
    LVL_IND,
    DataVariable,
    LinkedVariable,
    VariableInterface,
    VariableName,
)

__all__ = ["ObservationModel"]


@dataclass(frozen=True)
class ObservationModel:
    """
    Base class for valid observation models that may be used in probabilistic models (stateless).

    In particular, it provides data & linked variables regarding observations and their attachment to the model
    (the negative log-likelihood - nll - to be minimized).

    Parameters
    ----------
    name : :obj:`str`
        The name of observed variable (to name the data variable & attachment term related to this observation).
    getter : function :class:`.Dataset` -> :class:`.WeightedTensor`
        The way to retrieve the observed values from the :class:`.Dataset` (as a :class:`.WeightedTensor`):
        e.g. all values, subset of values - only x, y, z features, one-hot encoded features, ...
    dist : :class:`.SymbolicDistribution`
        The symbolic distribution, parametrized by model variables, for observed values (so to compute attachment).
    extra_vars : None (default) or Mapping[VarName, :class:`.VariableInterface`]
        Some new variables that are needed to fully define the symbolic distribution or the sufficient statistics.
        (e.g. "noise_std", and "y_L2_per_ft" for instance for a Gaussian model)
    """

    name: VariableName
    getter: Callable[[Dataset], WeightedTensor]
    dist: SymbolicDistribution
    extra_vars: Optional[TMapping[VariableName, VariableInterface]] = None

    def get_nll_attach_var_name(self, named_attach_vars: bool = True) -> str:
        """
        Return the name of the negative log likelihood attachement
        variable.
        """
        return f"nll_attach_{self.name}" if named_attach_vars else "nll_attach"

    def get_variables_specs(
        self,
        named_attach_vars: bool = True,
    ) -> dict[VariableName, VariableInterface]:
        """
        Automatic specifications of variables for this observation model.

        Parameters
        ----------
        named_attached_vars ::obj:`bool`, optional

        Returns
        -------
        :obj:`dict` [ :class:`~leaspy.variables.specs.VariableName`, :class:`~leaspy.variables.specs.VariableInterface`] 
            A dictionary mapping variable name to their correspondind specifications with
            - the primary DaraVariable
            - any `extra_vars` defined by the model
            - nll attachment variables :
                - nll_attach_var_ind: a :class:`~leaspy.variables.specs.LinkedVariable` representing the individual-level
                negative log-likelihood contributions
                - nll_attach_var: a :class:`~leaspy.variables.specs.LinkedVariable` that sums the individual contributions

        Notes
        -----
        The distribution object `self.dist`should provide a `get_func_nll(name)` method that
        returns a callable for computing the nll
        """
        # TODO change? a bit dirty? possibility of having aliases for variables?

        nll_attach_var = self.get_nll_attach_var_name(named_attach_vars)

        return {
            self.name: DataVariable(),
            # Dependent vars
            **(self.extra_vars or {}),
            # Attachment variables
            # not really memory efficient nor useful...
            # f"{nll_attach_var}_full": LinkedVariable(self.dist.get_func_nll(self.name)),
            f"{nll_attach_var}_ind": LinkedVariable(
                # SumDim(f"{nll_attach_var}_full", but_dim=LVL_IND)
                self.dist.get_func_nll(self.name).then(sum_dim, but_dim=LVL_IND)
            ),
            nll_attach_var: LinkedVariable(SumDim(f"{nll_attach_var}_ind")),
            # TODO jacobian of {nll_attach_var}_ind_jacobian_{self.name} wrt "y" as well? (for scipy minimize)
        }

    def serialized(self) -> Any:
        """
        Returns a JSON-exportable representation of the instance, excluding its name.

        Returns
        -------
        Any
            A representation of the instance, currently based on `repr(self.dist)`, 
            that is intended to be JSON-serializable.
        """
        # TODO: dirty for now to go fast
        return repr(self.dist)

    def to_dict(self) -> dict:
        """To be implemented..."""
        return {}

    def to_string(self) -> str:
        """
        Returns a string representation of the parameter for saving

        Returns
        -------
        :obj:`str`
            A string representation of the parameter, as stored in `self.string_for_json`.
        """
        return self.string_for_json



"""This module defines the `ModePosterior` sampler based personalize algorithm."""

import torch

from leaspy.utils.typing import DictParamsTorch

from ..base import AlgorithmName
from .mcmc import McmcPersonalizeAlgorithm

__all__ = ["ModePosteriorAlgorithm"]


class ModePosteriorAlgorithm(McmcPersonalizeAlgorithm):
    """Sampler-based algorithm that derives individual parameters as the most frequent mode posterior value from `n_iter` samplings.

    TODO? we could derive some confidence intervals on individual parameters thanks to this personalization algorithm...

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        Settings of the algorithm.
    """

    name: AlgorithmName = AlgorithmName.PERSONALIZE_MODE_POSTERIOR
    regularity_factor: float = 1.0
    """Weighting of regularity term in the final loss to be minimized."""

    def _compute_individual_parameters_from_samples_torch(
        self,
        values: DictParamsTorch,
        attachments: torch.Tensor,
        regularities: torch.Tensor,
    ) -> DictParamsTorch:
        """
        Compute dictionary of individual parameters from stacked values, attachments and regularities.

        Parameters
        ----------
        values : :obj:`dict`[ind_var_name: str, `torch.Tensor[float]` of shape (n_iter, n_individuals, *ind_var.shape)]
            The stacked history of values for individual latent variables.
        attachments : `torch.Tensor[float]` of shape (n_iter, n_individuals)
            The stacked history of attachments (per individual).
        regularities : `torch.Tensor[float]` of shape (n_iter, n_individuals)
            The stacked history of regularities (per individual; but summed on all individual variables and all of their dimensions).

        Returns
        -------
        :obj:`dict`[ind_var_name: :obj:`str`, `torch.Tensor[float]` of shape (n_individuals, *ind_var.shape)]
        """
        # Indices of iterations where loss (= negative log-likelihood) was minimal
        # (per individual, but tradeoff on ALL individual parameters)
        indices_iter_best = torch.argmin(
            attachments + self.regularity_factor * regularities, dim=0
        )  # shape (n_individuals,)
        indices_individuals = torch.arange(len(indices_iter_best))
        return {
            ind_var_name: value_var[indices_iter_best, indices_individuals]
            for ind_var_name, value_var in values.items()
        }



"""This module defines the `MeanPosterior` sampler based personalize algorithm."""

import torch

from leaspy.utils.typing import DictParamsTorch

from ..base import AlgorithmName
from .mcmc import McmcPersonalizeAlgorithm

__all__ = ["MeanPosteriorAlgorithm"]


class MeanPosteriorAlgorithm(McmcPersonalizeAlgorithm):
    """Sampler-based algorithm that derives individual parameters as the most frequent mean posterior value from `n_iter` samplings.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        Settings of the algorithm.
    """

    name: AlgorithmName = AlgorithmName.PERSONALIZE_MEAN_POSTERIOR

    def _compute_individual_parameters_from_samples_torch(
        self,
        values: DictParamsTorch,
        attachments: torch.Tensor,
        regularities: torch.Tensor,
    ) -> DictParamsTorch:
        """
        Compute dictionary of individual parameters from stacked values, attachments and regularities.

        Parameters
        ----------
        values : dict[ind_var_name: str, `torch.Tensor[float]` of shape (n_iter, n_individuals, *ind_var.shape)]
            The stacked history of values for individual latent variables.
        attachments : `torch.Tensor[float]` of shape (n_iter, n_individuals)
            The stacked history of attachments (per individual).
        regularities : `torch.Tensor[float]` of shape (n_iter, n_individuals)
            The stacked history of regularities (per individual; but summed on all individual variables and all of their dimensions).

        Returns
        -------
        dict[ind_var_name: str, `torch.Tensor[float]` of shape (n_individuals, *ind_var.shape)]
        """
        # Only compute the mean of posterior values (attachments & regularities not taken into account)
        return {
            ind_var_name: value_var.mean(dim=0)
            for ind_var_name, value_var in values.items()
        }



"""This module defines the `AbstractFitAlgo` class used for fitting algorithms."""

from typing import Optional

from leaspy.utils.typing import DictParamsTorch

from ..base import AlgorithmType, IterativeAlgorithm, ModelType, ReturnType
from ..settings import AlgorithmSettings, OutputsSettings

__all__ = ["FitAlgorithm"]


class FitAlgorithm(IterativeAlgorithm[ModelType, ReturnType]):
    r"""Abstract class containing common method for all `fit` algorithm classes.

    The algorithm is proven to converge if the sequence `burn_in_step` is positive, with an
    infinite sum :math:`\sum_k \epsilon_k = +\infty` and a finite sum of the squares
    :math:`\sum_k \epsilon_k^2 < \infty` (see following paper).

    `Construction of Bayesian Deformable Models via a Stochastic Approximation Algorithm: A Convergence Study <https://arxiv.org/abs/0706.0787>`_

    Parameters
    ----------
    settings : :class:`~leaspy.algo.AlgorithmSettings`
        The specifications of the algorithm as a :class:`~leaspy.algo.AlgorithmSettings` instance.

    Attributes
    ----------
    algorithm_device : :obj:`str`
        Valid :class:`torch.device`
    current_iteration : :obj:`int`, default 0
        The number of the current iteration.
        The first iteration will be 1 and the last one `n_iter`.
    sufficient_statistics : :obj:`dict` [:obj:`str`, :class:`torch.Tensor`] or None
        Sufficient statistics of the previous step.
        It is None during all the burn-in phase.
    output_manager : :class:`~leaspy.io.logs.fit_output_manager.FitOutputManager`
        Optional output manager of the algorithm
    Inherited attributes
        From :class:`~leaspy.algo.AbstractAlgo`

    See Also
    --------
    :meth:`leaspy.api.Leaspy.fit`
    """

    family = AlgorithmType.FIT

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.logs = settings.logs
        self.sufficient_statistics: Optional[DictParamsTorch] = None

    def set_output_manager(self, output_settings: OutputsSettings) -> None:
        """Set a :class:`~leaspy.algo.fit.FitOutputManager` object for the run of the algorithm.

        Parameters
        ----------
        output_settings : :class:`~leaspy.algo.OutputsSettings`
            Contains the logs settings for the computation run (console print periodicity, plot periodicity ...)

        Examples
        --------
        >>> from leaspy.algo import AlgorithmSettings, algorithm_factory, OutputsSettings
        >>> algo_settings = AlgorithmSettings("mcmc_saem")
        >>> my_algo = algorithm_factory(algo_settings)
        >>> settings = {
            'path': 'brouillons',
            'print_periodicity': 50,
            'plot_periodicity': 100,
            'save_periodicity': 50
        }
        >>> my_algo.set_output_manager(OutputsSettings(settings))
        """
        if output_settings is not None:
            from .fit_output_manager import FitOutputManager

            self.output_manager = FitOutputManager(output_settings)

    def _get_fit_metrics(self) -> Optional[dict[str, float]]:
        # TODO: finalize metrics handling, a bit dirty to place them in sufficient stats, only with a prefix...
        if self.sufficient_statistics is None:
            return None
        return {
            # (scalars only)
            k: v.item()
            for k, v in self.sufficient_statistics.items()
            if k.startswith("nll_")
        }

    def __str__(self) -> str:
        out = super().__str__()
        # add the fit metrics after iteration number (included the sufficient statistics for now...)
        fit_metrics = self._get_fit_metrics() or {}
        if len(fit_metrics):
            out += "\n= Metrics ="
            for m, v in fit_metrics.items():
                out += f"\n    {m} : {v:.5g}"
        return out



import warnings

import numpy as np
import statsmodels.api as sm
import torch
from statsmodels.regression.mixed_linear_model import MixedLM, MixedLMParams

from leaspy.exceptions import LeaspyDataInputError
from leaspy.io.data import Dataset
from leaspy.models import LMEModel

from ..base import AlgorithmName
from ..settings import AlgorithmSettings, OutputsSettings
from .base import FitAlgorithm

__all__ = ["LMEFitAlgorithm"]


class LMEFitAlgorithm(FitAlgorithm[LMEModel, tuple]):
    """Fitting algorithm associated to :class:`~.models.LMEModel`.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        * force_independent_random_effects : :obj:`bool`
            Force independence of random intercept & random slope
        * other keyword arguments passed to :meth:`statsmodels.regression.mixed_linear_model.MixedLM.fit`

    See Also
    --------
    :class:`statsmodels.regression.mixed_linear_model.MixedLM`
    """

    name: AlgorithmName = AlgorithmName.FIT_LME

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        params = settings.parameters.copy()
        # Algorithm true parameters
        self.force_independent_random_effects = params.pop(
            "force_independent_random_effects"
        )
        # Remaining parameters are parameters of statsmodels `fit` method
        self.sm_fit_parameters = params  # popped from other params

    def _run(self, model: LMEModel, dataset: Dataset, **kwargs) -> tuple:
        """Main method, refer to abstract definition in :meth:`~.algo.fit.abstract_fit_algo.AbstractFitAlgo.run`.

        Parameters
        ----------
        model : :class:`~leaspy.models.LMEModel`
            The used model.

        dataset : :class:`.Dataset`
            Dataset object

        Returns
        -------
        2-tuple:
            * None
            * noise scale (std-dev), scalar
        """
        if model.dimension != len(dataset.headers):
            raise ValueError("LME is only univariate")

        # get data
        ages = self._get_reformated(dataset, "timepoints")
        ages_mean, ages_std = np.mean(ages).item(), np.std(ages).item()
        ages_norm = (ages - ages_mean) / ages_std

        y = self._get_reformated(dataset, "values")
        subjects_with_repeat = self._get_reformated_subjects(dataset)

        # model
        X = sm.add_constant(ages_norm, prepend=True, has_constant="add")

        if model.with_random_slope_age:
            exog_re = X

            if self.force_independent_random_effects:
                free = MixedLMParams.from_components(
                    fe_params=np.ones(2), cov_re=np.eye(2)
                )
                self.sm_fit_parameters["free"] = free
                methods_not_compat_with_free = {"powell", "nm"}.intersection(
                    self.sm_fit_parameters["method"]
                )  # cf. statsmodels doc
                if len(methods_not_compat_with_free) > 0:
                    warnings.warn(
                        "<!> Methods {'powell','nm'} are not compatible with `force_independent_random_effects`"
                    )
        else:
            exog_re = None  # random_intercept only

        lme = MixedLM(y, X, subjects_with_repeat, exog_re, missing="raise")
        fitted_lme = lme.fit(**self.sm_fit_parameters)

        try:
            cov_re_unscaled_inv = np.linalg.inv(fitted_lme.cov_re_unscaled)
        except np.linalg.LinAlgError:
            raise LeaspyDataInputError(
                "Cannot predict random effects from " "singular covariance structure."
            )

        parameters = {
            "ages_mean": ages_mean,
            "ages_std": ages_std,
            "fe_params": fitted_lme.fe_params,
            "cov_re": fitted_lme.cov_re,
            "cov_re_unscaled_inv": cov_re_unscaled_inv,  # really useful for personalization
            "noise_std": fitted_lme.scale**0.5,  # statsmodels scale is variance
            "bse_fe": fitted_lme.bse_fe,
            "bse_re": fitted_lme.bse_re,
        }

        # update model parameters
        model.load_parameters(parameters)

        # return `(fitted_lme.resid ** 2).mean() ** .5` instead of scale?
        return None, parameters["noise_std"]

    @staticmethod
    def _get_reformated(dataset: Dataset, elem):
        # reformat ages
        dataset_elem = getattr(dataset, elem)
        # flatten
        flat_elem = torch.flatten(dataset_elem).numpy()
        # remove padding & nans
        final_elem = flat_elem[torch.flatten(dataset.mask > 0)]
        return final_elem

    @staticmethod
    def _get_reformated_subjects(dataset: Dataset):
        subjects_with_repeat = []
        for ind, subject in enumerate(dataset.indices):
            subjects_with_repeat += [subject] * max(
                dataset.n_visits_per_individual
            )  # [ind]
        subjects_with_repeat = np.array(subjects_with_repeat)
        # remove padding & nans
        subjects_with_repeat = subjects_with_repeat[torch.flatten(dataset.mask > 0)]
        return subjects_with_repeat

    def set_output_manager(self, output_settings: OutputsSettings) -> None:
        """
        Not implemented.
        """
        if output_settings is not None:
            warnings.warn("Settings logs in lme fit algorithm is not supported.")
        return



from ._base import ObservationModel
from ._bernoulli import BernoulliObservationModel
from ._factory import (
    OBSERVATION_MODELS,
    ObservationModelFactoryInput,
    ObservationModelNames,
    observation_model_factory,
)
from ._gaussian import FullGaussianObservationModel, GaussianObservationModel
from ._weibull import (
    AbstractWeibullRightCensoredObservationModel,
    WeibullRightCensoredObservationModel,
    WeibullRightCensoredWithSourcesObservationModel,
)

__all__ = [
    "BernoulliObservationModel",
    "FullGaussianObservationModel",
    "GaussianObservationModel",
    "ObservationModel",
    "ObservationModelFactoryInput",
    "ObservationModelNames",
    "observation_model_factory",
    "OBSERVATION_MODELS",
    "AbstractWeibullRightCensoredObservationModel",
    "WeibullRightCensoredObservationModel",
    "WeibullRightCensoredWithSourcesObservationModel",
]
